# -*- coding: utf-8 -*-
"""
Created on Fri Apr 12 09:56:22 2024

@author: Folkers

Function for the calculation of the efficiency of a radial compressor through 
two dimensional loss correlations. Necessary input alligns with the output of 
the transfer function of geometry for compressors.

Calculations of the thermodynamical properties are done using REFPROP. 
"""

"Module Import"
import numpy as np
import matplotlib.pyplot as plt
import math
from scipy.optimize import fsolve
from CompressorRadialVelocityTriangles import CompressorRadialVelocityTriangles

# Import of the REFPROP functions
import REFPROP2Py as REFPROP
Units = 21 # Mass based units

# Import of the transfer function of geometry
from CompressorGeometryAnalytical import CompressorGeometry

# Number of Iterations
Iterations = 3



"_____________________________________________________________________________"
"Section for the stand-alone use of this transfer function"
"Compression Task Definition"
# Working fluid in REFPROP nomenclature
WorkingFluid = "co2"
WorkingFluidComposition = [1]
# Volume flow at inlet conditions [m^3/s]
InletVolumeflow = 10
### Inlet Conditions ###
# Inlet temperature [K]
InletTemperature = 300
# Inlet pressure [kPa]
InletPressure = 100000
# Calculated values
InletEnthalpy = REFPROP.refprop(WorkingFluid, "TP", "H", Units, 0, 0, InletTemperature, InletPressure, WorkingFluidComposition).Output[0]
InletEntropy = REFPROP.refprop(WorkingFluid, "TP", "S", Units, 0, 0, InletTemperature, InletPressure, WorkingFluidComposition).Output[0]
InletDensity = REFPROP.refprop(WorkingFluid, "TP", "D", Units, 0, 0, InletTemperature, InletPressure, WorkingFluidComposition).Output[0]
InletSpeedOfSound = REFPROP.refprop(WorkingFluid, "TP", "W", Units, 0, 0, InletTemperature, InletPressure, WorkingFluidComposition).Output[0]

### Outlet Conditions ###
# Outlet pressure [kPa]
OutletPressure = 1000000

def RadialCompressorDesignFunction(WorkingFluid, WorkingFluidComposition, InletPressure, InletTemperature, OutletPressure, VolumeFlow):
    """
    Function for the design of radial compressors. The basic design parameters
    are chosen through the inverse method based on the Cordier diagram. 
    Afterwards the efficiency of the designed compressor is calculated 
    iteratively through one-dimensional loss correlations.
    
    If the program determines an axial machine to be more appropriate for the
    given compression task, an error occurs.
    
    INPUT: 
        - The working fluid or mixture as appicable to REFPROP
        - The working fluid composition as mass fractions
        - The inlet pressure in Pa
        - The inlet temperature in K
        - The outlet pressure in Pa
        - The inlet volume flow in m^3/s
        
    OUTPUT: A dictionary containing the following variables in SI units
        - Isentropic Efficiency
        - Total Efficiency
        - Static Efficiency
        - Inlet Tip Diameter
        - Inlet Hub Diameter
        - Rotor Outlet Diameter
        - Stator Inlet Diameter
        - Stator Outlet Diameter
        - Rotor Blade Number
        - Stator Blade Number
        - Rotor Axial Lenght
        - Inlet Velocity
        - Outlet Velocity
        - Outlet Temperature
        - Outlet Pressure
        - Outlet Enthalpy
        - Outlet Entropy
        - Outlet Density
        - Stage Number
        - Work Coefficient
        - Flow Coefficient
        - Specific Speed
        - Specific Diameter
    """
    
    
    "Functions"
    def SlipFunctionRotor(x):
        # Function for calculating the slip factor, rotor stage number and relative
        # blade angles at rotor outlet
        # x[0]: Slip Factor
        # x[1]: Relative Blade Angle 2
        # x[2]: Rotor Blade Number
        equation1 = x[0] - ( 1-(math.cos(x[1]))**0.5 / x[2]**0.7 )
        equation2 = x[1] - ( math.atan(1/(FlowCoefficient*VarOutput[2])-math.tan(Alpha2/x[0])) )
        equation3 = x[2] - ( 2*np.pi*math.cos((Beta1MeanDiameter[i]+x[1])/2)/(0.4*math.log(1/TipDiameterRatio[i])) )
        return equation1, equation2, equation3
    
    def SkinFrictionFunctionRotor(x):
        equation = 1/x**0.5 - (-4*math.log10(1.255/(RotorSkinFrictionReynoldsNumber[i]*x**0.5)))
        return equation
    
    def SkinFrictionFunctionRotor4000(x):
        equation = 1/x**0.5 - (-4*math.log10(1.255/(4000*x**0.5)))
        return equation
    
    def SkinFrictionFunctionStatorVaneless(x):
        equation = 1/x**0.5 - (-4*math.log10(1.255/(StatorVanelessSkinFrictionReynoldsNumber[i]*x**0.5)))
        return equation
    
    def SkinFrictionFunctionStatorVaneless4000(x):
        equation = 1/x**0.5 - (-4*math.log10(1.255/(4000*x**0.5)))
        return equation   
     
    def SkinFrictionFunctionStatorVaned(x):
        equation = 1/x**0.5 - (-4*math.log10(1.255/(StatorVanedSkinFrictionReynoldsNumber[i]*x**0.5)))
        return equation
    
    def SkinFrictionFunctionStatorVaned4000(x):
        equation = 1/x**0.5 - (-4*math.log10(1.255/(4000*x**0.5)))
        return equation
    # Calculated values
    IsentropicOutletEnthalpy = REFPROP.refprop(WorkingFluid, "SP", "H", Units, 0, 0, InletEntropy, OutletPressure, WorkingFluidComposition).Output[0]
    
    # Isentropic specific work over the machine  
    SpecificWork = IsentropicOutletEnthalpy-InletEnthalpy
    
    # Calling the transfer function of geometry
    MachineType, StageNumber, SpecificDiameter, SpecificSpeed, EstimatedTotalToStaticEfficiency, SizeCoefficient, Speed, OuterDiameter, WorkCoefficient, FlowCoefficient, AbsoluteVelocity, RelativeVelocity, RotationalVelocity, Alpha, VarOutput, Beta, MinimumOutletDiameter, MinimumStageNumber = CompressorGeometry(SpecificWork, InletVolumeflow, InletSpeedOfSound)
    
    # Check to ensure a radial machine is being designed
    if MachineType == "Axial":
        # raise ValueError("The input parameters resulted in an axial compressor. Consider decreasing the volume flow or increasing the specific work to arrive at a radial compressor.")
        ResultDictionary = {"Isentropic Efficiency": None,
                            "Total Efficiency": None,
                            "Static Efficiency": None,
                            "Inlet Tip Diameter": None,
                            "Inlet Hub Diameter": None,
                            "Rotor Outlet Diameter": None,
                            "Stator Inlet Diameter": None,
                            "Stator Outlet Diameter": None,
                            "Rotor Blade Number": None,
                            "Stator Blade Number": None,
                            "Rotor Axial Lenght": None,
                            "Inlet Velocity": None,
                            "Outlet Velocity": None,
                            "Outlet Temperature": None,
                            "Outlet Pressure": None,
                            "Outlet Enthalpy": None,
                            "Outlet Entropy": None,
                            "Outlet Density": None,
                            "Stage Number": None,
                            "Work Coefficient": None,
                            "Flow Coefficient": None,
                            "Specific Speed": None,
                            "Specific Diameter": None}
        print("_____________________________________________________________________________")
        print("The input parameters resulted in an axial compressor. Consider decreasing the volume flow or increasing the specific work to arrive at a radial compressor. The output is given as an empty dictionary.")
        print("_____________________________________________________________________________")
        return ResultDictionary
    
    # Rough estimation of the isentropic efficiency to use as a starting point
    EstimatedIsentropicEfficiency = np.zeros(StageNumber)+SpecificWork/StageNumber / ((SpecificWork/StageNumber+AbsoluteVelocity[0]**2/2)/EstimatedTotalToStaticEfficiency)
    OverallEstimatedIsentropicEfficiency = SpecificWork/StageNumber / ((SpecificWork/StageNumber+AbsoluteVelocity[0]**2/2)/EstimatedTotalToStaticEfficiency)
    EstimatedIsentropicRotorEfficiency = np.zeros(StageNumber)+EstimatedIsentropicEfficiency**0.5  # EstimatedIsentropicEfficiency
    EstimatedIsentropicStatorEfficiency = np.zeros(StageNumber)+EstimatedIsentropicEfficiency**0.5
    # Constant stage pressure ratio
    PressureRatio = (OutletPressure/InletPressure)**(1/StageNumber)
    Pressure3 = np.zeros(StageNumber)
    Pressure3[0] = InletPressure*PressureRatio
    for i in range(1,StageNumber):
        Pressure3[i] = Pressure3[i-1]*PressureRatio
    
    Alpha1 = Alpha[0]
    Alpha2 = Alpha[1]
    Beta1 = Beta[0]
    Beta2 = Beta[1]
    RotationalVelocity1 = RotationalVelocity[0]
    RotationalVelocity2 = RotationalVelocity[1]
    AbsoluteVelocity1 = AbsoluteVelocity[0]
    AbsoluteVelocity2 = AbsoluteVelocity[1]
    RelativeVelocity1 = RelativeVelocity[0]
    RelativeVelocity2 = RelativeVelocity[1]
    AbsoluteMeridionalVelocity1 = AbsoluteVelocity1*math.cos(Alpha1)
    AbsoluteRotationalVelocity1 = AbsoluteVelocity1*math.sin(Alpha1)
    AbsoluteMeridionalVelocity2 = AbsoluteVelocity2*math.cos(Alpha2)
    AbsoluteRotationalVelocity2 = AbsoluteVelocity2*math.sin(Alpha2)
    RelativeMeridionalVelocity1 = RelativeVelocity1*math.cos(Beta1)
    RelativeRotationalVelocity1 = RelativeVelocity1*math.sin(Beta1)
    RelativeMeridionalVelocity2 = RelativeVelocity2*math.cos(Beta2)
    RelativeRotationalVelocity2 = RelativeVelocity2*math.sin(Beta2)
    
    "_____________________________________________________________________________"
    "Setting of Variables"
    # Variables for the thermodynamic properties
    Enthalpy1 = np.zeros(StageNumber)
    Entropy1 = np.zeros(StageNumber)
    Pressure1 = np.zeros(StageNumber)
    Temperature1 = np.zeros(StageNumber)
    Density1 = np.zeros(StageNumber)
    Enthalpy2 = np.zeros(StageNumber)
    Entropy2 = np.zeros(StageNumber)
    Pressure2 = np.zeros(StageNumber)
    Temperature2 = np.zeros(StageNumber)
    Density2 = np.zeros(StageNumber)
    Enthalpy3 = np.zeros(StageNumber)
    Entropy3 = np.zeros(StageNumber)
    # Pressure3 = np.zeros(StageNumber)
    Temperature3 = np.zeros(StageNumber)
    Density3 = np.zeros(StageNumber)
    TotalEnthalpy1 = np.zeros(StageNumber)
    TotalEntropy1 = np.zeros(StageNumber)
    TotalPressure1 = np.zeros(StageNumber)
    TotalTemperature1 = np.zeros(StageNumber)
    TotalDensity1 = np.zeros(StageNumber)
    TotalEnthalpy2 = np.zeros(StageNumber)
    TotalEntropy2 = np.zeros(StageNumber)
    TotalPressure2 = np.zeros(StageNumber)
    TotalTemperature2 = np.zeros(StageNumber)
    TotalDensity2 = np.zeros(StageNumber)
    TotalEnthalpy3 = np.zeros(StageNumber)
    TotalEntropy3 = np.zeros(StageNumber)
    TotalPressure3 = np.zeros(StageNumber)
    TotalTemperature3 = np.zeros(StageNumber)
    TotalDensity3 = np.zeros(StageNumber)
    TotalRelativeEnthalpy1 = np.zeros(StageNumber)
    TotalRelativeRelativeEntropy1 = np.zeros(StageNumber)
    TotalRelativePressure1 = np.zeros(StageNumber)
    TotalRelativeTemperature1 = np.zeros(StageNumber)
    TotalRelativeDensity1 = np.zeros(StageNumber)
    TotalRelativeEnthalpy2 = np.zeros(StageNumber)
    TotalRelativeEntropy2 = np.zeros(StageNumber)
    TotalRelativePressure2 = np.zeros(StageNumber)
    TotalRelativeTemperature2 = np.zeros(StageNumber)
    TotalRelativeDensity2 = np.zeros(StageNumber)
    TotalRelativeEnthalpy3 = np.zeros(StageNumber)
    TotalRelativeEntropy3 = np.zeros(StageNumber)
    TotalRelativePressure3 = np.zeros(StageNumber)
    TotalRelativeTemperature3 = np.zeros(StageNumber)
    TotalRelativeDensity3 = np.zeros(StageNumber)
    IsentropicEnthalpy2 = np.zeros(StageNumber)
    IsentropicEnthalpy3 = np.zeros(StageNumber)
    Rothalpy1 = np.zeros(StageNumber)
    Rothalpy2 = np.zeros(StageNumber)
    SpeedOfSound1 = np.zeros(StageNumber)
    SpeedOfSound2 = np.zeros(StageNumber)
    SpeedOfSound3 = np.zeros(StageNumber)
    Viscosity1 = np.zeros(StageNumber)
    Viscosity2 = np.zeros(StageNumber)
    Viscosity3 = np.zeros(StageNumber)
    AbsoluteVelocity3 = np.zeros(StageNumber)
    LossTemperature2 = np.zeros(StageNumber)
    LossTemperature3 = np.zeros(StageNumber)
    LossPressure2 = np.zeros(StageNumber)
    LossPressure3 = np.zeros(StageNumber)
    LossTotalEnthalpy3Alt = np.zeros(StageNumber)
    
    # Variables for the Geometric Parameters
    TipDiameter1 = np.zeros(StageNumber)
    TipDiameterRatio = np.zeros(StageNumber)
    HubDiameter1 = np.zeros(StageNumber)
    Diameter2 = np.zeros(StageNumber)
    Volumeflow1 = np.zeros(StageNumber)
    Volumeflow2 = np.zeros(StageNumber)
    Volumeflow3 = np.zeros(StageNumber)
    FlowArea1 = np.zeros(StageNumber)
    FlowArea2 = np.zeros(StageNumber)
    FlowArea3 = np.zeros(StageNumber)
    BladeHeight2 = np.zeros(StageNumber)
    HubDiameter1 = np.zeros(StageNumber)
    BladeHeight1 = np.zeros(StageNumber)
    HubDiameterRatio = np.zeros(StageNumber)
    MeanDiameter1 = np.zeros(StageNumber)
    Alpha1MeanDiameter = np.zeros(StageNumber)
    Beta1MeanDiameter = np.zeros(StageNumber)
    MeanRelativeVelocity1 = np.zeros(StageNumber)
    SlipFactor = np.zeros(StageNumber)
    RotorBladeNumber = np.zeros(StageNumber)
    Beta2Blade = np.zeros(StageNumber)
    BetaMean = np.zeros(StageNumber)
    RotorBladeThickness = np.zeros(StageNumber)
    AxialClearence = np.zeros(StageNumber)
    RadialClearence = np.zeros(StageNumber)
    BackfaceClearence = np.zeros(StageNumber)
    RotorInletBladePitch = np.zeros(StageNumber)
    RotorOutletBladePitch = np.zeros(StageNumber)
    RotorInletBladeDistance = np.zeros(StageNumber)
    RotorOutletBladeDistance = np.zeros(StageNumber)
    RotorInletCrossSection = np.zeros(StageNumber)
    RotorOutletCrossSection = np.zeros(StageNumber)
    RotorInletWettedPerimeter = np.zeros(StageNumber)
    RotorOutletWettedPerimeter = np.zeros(StageNumber)
    RotorInletHydraulicDiameter = np.zeros(StageNumber)
    RotorOutletHydraulicDiameter = np.zeros(StageNumber)
    RotorHydraulicDiameter = np.zeros(StageNumber) 
    RotorAxialLength = np.zeros(StageNumber)
    RotorMeridionalLength = np.zeros(StageNumber)
    RotorHydraulicLength = np.zeros(StageNumber)
    StatorAlpha2 = np.zeros(StageNumber)
    StatorInletMachNumber = np.zeros(StageNumber)
    StatorDiameter2 = np.zeros(StageNumber)
    RotorOutletAbsoluteRotationalVelocity = np.zeros(StageNumber)
    StatorInletAbsoluteRotationalVelocity = np.zeros(StageNumber)
    StatorInletAbsoluteMeridionalVelocity = np.zeros(StageNumber)
    StatorInletAbsoluteVelocity = np.zeros(StageNumber)
    StatorInletBladeHeight = np.zeros(StageNumber)
    StatorVanelessHydraulicLength = np.zeros(StageNumber)
    StatorVanelessHydraulicDiameter = np.zeros(StageNumber)
    StatorOutletDiameter = np.zeros(StageNumber)
    StatorOutletBladeHeight = np.zeros(StageNumber)
    StatorOutletAbsoluteMeridionalVelocity = np.zeros(StageNumber)
    StatorOutletAbsoluteVelocity = np.zeros(StageNumber)
    StatorBladeNumber = np.zeros(StageNumber)
    StatorVanedMeridionalLength = np.zeros(StageNumber)
    StatorVanedHydraulicLength = np.zeros(StageNumber)
    StatorInletBladePitch = np.zeros(StageNumber)
    StatorOutletBladePitch = np.zeros(StageNumber)
    StatorSolidity = np.zeros(StageNumber)
    StatorInletBladeDistance = np.zeros(StageNumber)
    StatorOutletBladeDistance = np.zeros(StageNumber)
    StatorInletCrossSection = np.zeros(StageNumber)
    StatorOutletCrossSection = np.zeros(StageNumber)
    StatorInletWettedPerimeter = np.zeros(StageNumber)
    StatorOutletWettedPerimeter = np.zeros(StageNumber)
    StatorVanedInletHydraulicDiameter = np.zeros(StageNumber)
    StatorVanedOutletHydraulicDiameter = np.zeros(StageNumber)
    StatorVanedHydraulicDiameter = np.zeros(StageNumber)
    RotorInletReynoldsNumber = np.zeros(StageNumber)
    RotorOutletReynoldsNumber = np.zeros(StageNumber)
    StatorInletReynoldsNumber = np.zeros(StageNumber)
    StatorOutletReynoldsNumber = np.zeros(StageNumber)
    RotorRoughness = np.zeros(StageNumber)
    StatorRoughness = np.zeros(StageNumber)
    StatorBladeThickness = np.zeros(StageNumber)
    
    
    # Variables for the Losses
    IncidenceLossCoefficient = np.zeros(StageNumber)
    MeanRelativeVelocity1 = np.zeros(StageNumber)
    MeanRelativeVelocity = np.zeros(StageNumber)
    RotorSkinFrictionReynoldsNumber = np.zeros(StageNumber)
    RotorSkinFrictionCoefficient = np.zeros(StageNumber)
    RotorSkinFrictionCoefficientSmooth = np.zeros(StageNumber)
    RotorSkinFrictionCoefficientRough = np.zeros(StageNumber)
    RotorSkinFrictionRoughReynoldsNumber = np.zeros(StageNumber)
    RotorSkinFrictionLossCoefficient = np.zeros(StageNumber)
    BladeLoadingLossCoefficient = np.zeros(StageNumber)
    Km = np.zeros(StageNumber)
    MeanBladeHeight = np.zeros(StageNumber)
    MeanRelativeVelocityAlt = np.zeros(StageNumber)
    HubToShroudLossCoefficient = np.zeros(StageNumber)
    EquivalentDiffusionFactor = np.zeros(StageNumber)
    SeperationVelocity = np.zeros(StageNumber)
    AbsoluteMeridionalMixingVelocity = np.zeros(StageNumber)
    AbsoluteMeridionalWakeVelocity = np.zeros(StageNumber)
    MixingLossCoefficient = np.zeros(StageNumber)
    ClearenceLossCoefficient = np.zeros(StageNumber)
    RotorLossCoefficient = np.zeros(StageNumber)
    
    AbsoluteQuadraticMeanVelocity = np.zeros(StageNumber)
    DivergenceParameter = np.zeros(StageNumber)
    ReferenceParameter = np.zeros(StageNumber)
    DiffusionEfficiency = np.zeros(StageNumber)
    StatorVanelessDiffusionLossCoefficient = np.zeros(StageNumber)
    StatorVanelessSkinFrictionReynoldsNumber = np.zeros(StageNumber)
    StatorVanelessSkinFrictionCoefficient = np.zeros(StageNumber)
    StatorVanelessSkinFrictionCoefficientSmooth = np.zeros(StageNumber)
    StatorVanelessSkinFrictionCoefficientRough = np.zeros(StageNumber)
    StatorVanelessSkinFrictionRoughReynoldsNumber = np.zeros(StageNumber)
    StatorVanelessSkinFrictionLossCoefficient = np.zeros(StageNumber)
    StatorVanelessLossCoefficient = np.zeros(StageNumber)
    
    StatorVanedIncidenceLossCoefficient = np.zeros(StageNumber)
    AbsoluteMeridionalMixingVelocity3 = np.zeros(StageNumber)
    StatorEquivalentDiffusionFactor = np.zeros(StageNumber)
    SeperationVelocity3 = np.zeros(StageNumber)
    AbsoluteMeridionalWakeVelocity3 = np.zeros(StageNumber)
    StatorVanedMixingLossCoefficient = np.zeros(StageNumber)
    StatorVanedSkinFrictionReynoldsNumber = np.zeros(StageNumber)
    StatorVanedSkinFrictionCoefficient = np.zeros(StageNumber)
    StatorVanedSkinFrictionCoefficientSmooth = np.zeros(StageNumber)
    StatorVanedSkinFrictionCoefficientRough = np.zeros(StageNumber)
    StatorVanedSkinFrictionRoughReynoldsNumber = np.zeros(StageNumber)
    A = np.zeros(StageNumber)
    StatorVanedSkinFrictionLossCoefficient = np.zeros(StageNumber)
    StatorVanedLossCoefficient = np.zeros(StageNumber)
    StatorLossCoefficient = np.zeros(StageNumber)
    
    # Variables for the Efficiencies
    IsentropicTotalRelativeEnthalpy2 = np.zeros(StageNumber)
    IsentropicTotalRelativePressure2 = np.zeros(StageNumber)
    LossTotalRelativePressure2 = np.zeros(StageNumber)
    LossEnthalpy2 = np.zeros(StageNumber)
    LossEntropy2 = np.zeros(StageNumber)
    RotorEfficiency = np.zeros(StageNumber)
    LossTotalEnthalpy2 = np.zeros(StageNumber)
    LossTotalPressure2 = np.zeros(StageNumber)
    LossPressure2 = np.zeros(StageNumber)
    LossStatorInletTotalPressure2 = np.zeros(StageNumber)
    LossTotalPressure3 = np.zeros(StageNumber)
    LossTotalEnthalpy3 = np.zeros(StageNumber)
    LossEntropy3 = np.zeros(StageNumber)
    LossPressure3 = np.zeros(StageNumber)
    LossEnthalpy3 = np.zeros(StageNumber)
    TotalEfficiency = np.zeros(StageNumber)
    StaticEfficiency = np.zeros(StageNumber)
    IsentropicEfficiency = np.zeros(StageNumber)
    StatorEfficiency = np.zeros(StageNumber)
    
    CalculatedIsentropicEfficiency = np.zeros(Iterations)
    
    "_____________________________________________________________________________"
    "Thermodynamic Inlet Properties"
    
    
    # Applying the inlet conditions
    Enthalpy1[0] = InletEnthalpy
    Entropy1[0] = InletEntropy
    Pressure1[0] = InletPressure
    Temperature1[0] = InletTemperature
    Density1[0] = InletDensity
    
    
    "Loop for redoing the calculations and updating the efficiency estimations"
    for j in range(Iterations):
        Alpha, Beta, RotationalVelocity, AbsoluteVelocity, RelativeVelocity, WorkCoefficient, FlowCoefficient, VarOutput = CompressorRadialVelocityTriangles(SpecificSpeed, Speed, OuterDiameter, SpecificWork/StageNumber/OverallEstimatedIsentropicEfficiency, InletVolumeflow, 0)
        Alpha1 = Alpha[0]
        Alpha2 = Alpha[1]
        Beta1 = Beta[0]
        Beta2 = Beta[1]
        RotationalVelocity1 = RotationalVelocity[0]
        RotationalVelocity2 = RotationalVelocity[1]
        AbsoluteVelocity1 = AbsoluteVelocity[0]
        AbsoluteVelocity2 = AbsoluteVelocity[1]
        RelativeVelocity1 = RelativeVelocity[0]
        RelativeVelocity2 = RelativeVelocity[1]
        AbsoluteMeridionalVelocity1 = AbsoluteVelocity1*math.cos(Alpha1)
        AbsoluteRotationalVelocity1 = AbsoluteVelocity1*math.sin(Alpha1)
        AbsoluteMeridionalVelocity2 = AbsoluteVelocity2*math.cos(Alpha2)
        AbsoluteRotationalVelocity2 = AbsoluteVelocity2*math.sin(Alpha2)
        RelativeMeridionalVelocity1 = RelativeVelocity1*math.cos(Beta1)
        RelativeRotationalVelocity1 = RelativeVelocity1*math.sin(Beta1)
        RelativeMeridionalVelocity2 = RelativeVelocity2*math.cos(Beta2)
        RelativeRotationalVelocity2 = RelativeVelocity2*math.sin(Beta2)
    
        for i in range(StageNumber):
            # Calculations of the thermodynamic properties stage by stage
            TotalEnthalpy1[i] = Enthalpy1[i]+AbsoluteVelocity1**2/2 
            TotalPressure1[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, TotalEnthalpy1[i], Entropy1[i], WorkingFluidComposition).Output[0]
            TotalTemperature1[i] = REFPROP.refprop(WorkingFluid, "HS", "T", Units, 0, 0, TotalEnthalpy1[i], Entropy1[i], WorkingFluidComposition).Output[0]
            TotalDensity1[i] = REFPROP.refprop(WorkingFluid, "HS", "D", Units, 0, 0, TotalEnthalpy1[i], Entropy1[i], WorkingFluidComposition).Output[0]
            TotalRelativeEnthalpy1[i] = Enthalpy1[i]+RelativeVelocity1**2/2 
            TotalRelativePressure1[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, TotalRelativeEnthalpy1[i], Entropy1[i], WorkingFluidComposition).Output[0]
            TotalRelativeTemperature1[i] = REFPROP.refprop(WorkingFluid, "HS", "T", Units, 0, 0, TotalRelativeEnthalpy1[i], Entropy1[i], WorkingFluidComposition).Output[0]
            
            Rothalpy1[i] = TotalRelativeEnthalpy1[i]-RotationalVelocity1**2/2 
            Rothalpy2[i] = Rothalpy1[i]
            Enthalpy2[i] = Rothalpy2[i]-RelativeVelocity2**2/2+RotationalVelocity2**2/2 
            IsentropicEnthalpy2[i] = Enthalpy1[i]+EstimatedIsentropicRotorEfficiency[i]*(Enthalpy2[i]-Enthalpy1[i])
            Pressure2[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, IsentropicEnthalpy2[i], Entropy1[i], WorkingFluidComposition).Output[0]
            Entropy2[i] = REFPROP.refprop(WorkingFluid, "HP", "S", Units, 0, 0, Enthalpy2[i], Pressure2[i], WorkingFluidComposition).Output[0]
            Temperature2[i] = REFPROP.refprop(WorkingFluid, "HP", "T", Units, 0, 0, Enthalpy2[i], Pressure2[i], WorkingFluidComposition).Output[0]    
            Density2[i] = REFPROP.refprop(WorkingFluid, "HP", "D", Units, 0, 0, Enthalpy2[i], Pressure2[i], WorkingFluidComposition).Output[0]
            TotalEnthalpy2[i] = Enthalpy2[i]+AbsoluteVelocity2**2/2 
            TotalPressure2[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, TotalEnthalpy2[i], Entropy2[i], WorkingFluidComposition).Output[0]
            TotalTemperature2[i] = REFPROP.refprop(WorkingFluid, "HS", "T", Units, 0, 0, TotalEnthalpy2[i], Entropy2[i], WorkingFluidComposition).Output[0]
            TotalDensity2[i] = REFPROP.refprop(WorkingFluid, "HS", "D", Units, 0, 0, TotalEnthalpy2[i], Entropy2[i], WorkingFluidComposition).Output[0]
            TotalRelativeEnthalpy2[i] = Enthalpy2[i]+RelativeVelocity[1]**2/2 
            TotalRelativePressure2[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, TotalRelativeEnthalpy2[i], Entropy2[i], WorkingFluidComposition).Output[0]
            TotalRelativeTemperature2[i] = REFPROP.refprop(WorkingFluid, "HS", "T", Units, 0, 0, TotalRelativeEnthalpy2[i], Entropy2[i], WorkingFluidComposition).Output[0]
            
            IsentropicEnthalpy3[i] = REFPROP.refprop(WorkingFluid, "PS", "H", Units, 0, 0, Pressure3[i], Entropy1[i], WorkingFluidComposition).Output[0]
            # TotalEnthalpy3[i] = TotalEnthalpy2[i]
            Enthalpy3[i] = Enthalpy1[i]+(IsentropicEnthalpy3[i]-Enthalpy1[i])/EstimatedIsentropicStatorEfficiency[i]
            # Enthalpy3[i] = TotalEnthalpy3[i]-AbsoluteVelocity1**2/2
            TotalEnthalpy3[i] = Enthalpy3[i]+AbsoluteVelocity1**2/2
            Entropy3[i] = REFPROP.refprop(WorkingFluid, "HP", "S", Units, 0, 0, Enthalpy3[i], Pressure3[i], WorkingFluidComposition).Output[0]
            Temperature3[i] = REFPROP.refprop(WorkingFluid, "HP", "T", Units, 0, 0, Enthalpy3[i], Pressure3[i], WorkingFluidComposition).Output[0]
            Density3[i] = REFPROP.refprop(WorkingFluid, "HP", "D", Units, 0, 0, Enthalpy3[i], Pressure3[i], WorkingFluidComposition).Output[0]
            TotalPressure3[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, TotalEnthalpy3[i], Entropy3[i], WorkingFluidComposition).Output[0]
            TotalTemperature3[i] = REFPROP.refprop(WorkingFluid, "HS", "T", Units, 0, 0, TotalEnthalpy3[i], Entropy3[i], WorkingFluidComposition).Output[0]
            TotalDensity3[i] = REFPROP.refprop(WorkingFluid, "HS", "D", Units, 0, 0, TotalEnthalpy3[i], Entropy3[i], WorkingFluidComposition).Output[0]
            
            SpeedOfSound1[i] = REFPROP.refprop(WorkingFluid, "TP", "W", Units, 0, 0, Temperature1[i], Pressure1[i], WorkingFluidComposition).Output[0]
            SpeedOfSound2[i] = REFPROP.refprop(WorkingFluid, "TP", "W", Units, 0, 0, Temperature2[i], Pressure2[i], WorkingFluidComposition).Output[0]
            SpeedOfSound3[i] = REFPROP.refprop(WorkingFluid, "TP", "W", Units, 0, 0, Temperature3[i], Pressure3[i], WorkingFluidComposition).Output[0]
            
            Viscosity1[i] = REFPROP.refprop(WorkingFluid, "TP", "Vis", Units, 0, 0, Temperature1[i], Pressure1[i], WorkingFluidComposition).Output[0]
            Viscosity2[i] = REFPROP.refprop(WorkingFluid, "TP", "Vis", Units, 0, 0, Temperature2[i], Pressure2[i], WorkingFluidComposition).Output[0]
            Viscosity3[i] = REFPROP.refprop(WorkingFluid, "TP", "Vis", Units, 0, 0, Temperature3[i], Pressure3[i], WorkingFluidComposition).Output[0]
            
            AbsoluteVelocity3[i] = AbsoluteVelocity1 # (2*(TotalEnthalpy3[i]-Enthalpy3[i]))**0.5
            
            # if i <StageNumber -1:
            #     Enthalpy1[i+1] = Enthalpy3[i]
            #     Entropy1[i+1] = Entropy3[i]
            #     Pressure1[i+1] = Pressure3[i]
            #     Temperature1[i+1] = Temperature3[i]
            #     Density1[i+1] = Density3[i]
            
                # End of for loop for thermodynamic properties
            
            "_____________________________________________________________________________"
            "Geometric Parameters"
            "Rotor Geoemtry"
            TipDiameter1[i] = OuterDiameter*VarOutput[0]
            TipDiameterRatio[i] = VarOutput[0]
            HubDiameter1[i] = OuterDiameter*VarOutput[1]
            Diameter2[i] = OuterDiameter
            
            Volumeflow1[i] = InletVolumeflow*InletDensity/Density1[i]
            Volumeflow2[i] = InletVolumeflow*InletDensity/Density2[i]
            Volumeflow3[i] = InletVolumeflow*InletDensity/Density3[i]
            
            FlowArea1[i] = Volumeflow1[i]/AbsoluteMeridionalVelocity1
            FlowArea2[i] = Volumeflow2[i]/AbsoluteMeridionalVelocity2
            FlowArea3[i] = Volumeflow3[i]/AbsoluteVelocity3[i]
        
            BladeHeight2[i] = FlowArea2[i]/np.pi/Diameter2[i]
            HubDiameter1[i]= (TipDiameter1[i]**2-4*FlowArea1[i]/np.pi)**0.5
            BladeHeight1[i] = (TipDiameter1[i]-HubDiameter1[i])/2
            HubDiameterRatio[i] = HubDiameter1[i]/Diameter2[i]
            MeanDiameter1[i] = (TipDiameter1[i]+HubDiameter1[i])/2
            Alpha1MeanDiameter[i] = math.atan(math.tan(Alpha1)*TipDiameter1[i]/HubDiameter1[i])
            Beta1MeanDiameter[i] = math.atan(TipDiameterRatio[i]/FlowCoefficient*MeanDiameter1[i]/TipDiameter1[i]-math.tan(Alpha1MeanDiameter[i]))
            MeanRelativeVelocity1[i] = math.atan(1/(2*FlowCoefficient)*(TipDiameterRatio[i]+HubDiameterRatio[i])-(2*TipDiameterRatio[i]*math.tan(Alpha1))/(TipDiameter1[i]+HubDiameter1[i]))
            SlipFactor[i], Beta2Blade[i], RotorBladeNumber[i] = fsolve(SlipFunctionRotor, [0.9, Beta2, 10])
            RotorBladeNumber[i] = np.floor(RotorBladeNumber[i])
            Beta2Blade[i] = 2*math.acos(0.2*RotorBladeNumber[i]*math.log(1/TipDiameterRatio[i])/np.pi)-Beta1MeanDiameter[i]
            if Beta2Blade[i] > np.pi/2:     # Check to ensure the values of Beta2Blade do not cause errors in the calculation of the slip factor
                Beta2Blade[i] = np.pi/2
            SlipFactor[i] = 1-(math.cos(Beta2Blade[i]))**0.5/RotorBladeNumber[i]**0.7
            BetaMean[i] = (Beta1MeanDiameter[i]+Beta2Blade[i])/2
            RotorBladeThickness[i] = 0.01*Diameter2[i]  # For shrouded impellers
            AxialClearence[i] = 0.05*BladeHeight2[i]
            RadialClearence[i] = 0.05*BladeHeight2[i]
            BackfaceClearence[i] = 0.05*BladeHeight2[i]
            RotorInletBladePitch[i] = np.pi*MeanDiameter1[i]/RotorBladeNumber[i]
            RotorOutletBladePitch[i] = np.pi*Diameter2[i]/RotorBladeNumber[i]
            RotorInletBladeDistance[i] = RotorInletBladePitch[i]*math.cos(Beta1MeanDiameter[i])
            RotorOutletBladeDistance[i] = RotorOutletBladePitch[i]*math.cos(Beta2Blade[i])
            RotorInletCrossSection[i] = RotorInletBladeDistance[i]*BladeHeight1[i]
            RotorOutletCrossSection[i] = RotorOutletBladeDistance[i]*BladeHeight2[i]
            RotorInletWettedPerimeter[i] = 2*(RotorInletBladeDistance[i]+BladeHeight1[i])
            RotorOutletWettedPerimeter[i] = 2*(RotorOutletBladeDistance[i]+BladeHeight2[i])
            RotorInletHydraulicDiameter[i] =2*RotorInletBladeDistance[i]*BladeHeight1[i]/(RotorInletBladeDistance[i]+BladeHeight1[i])
            RotorOutletHydraulicDiameter[i] =2*RotorOutletBladeDistance[i]*BladeHeight2[i]/(RotorOutletBladeDistance[i]+BladeHeight2[i])
            RotorHydraulicDiameter[i] = (RotorInletHydraulicDiameter[i]+RotorOutletHydraulicDiameter[i])/2 
            RotorAxialLength[i] = Diameter2[i]*(0.014+0.023/HubDiameterRatio[i]+1.58*(TipDiameterRatio[i]**2-HubDiameterRatio[i]**2)*FlowCoefficient)
            RotorMeridionalLength[i] = np.pi/8*(2*RotorAxialLength[i]-BladeHeight2[i]+Diameter2[i]-TipDiameter1[i]+BladeHeight1[i])
            RotorHydraulicLength[i] = RotorMeridionalLength[i]/math.cos(BetaMean[i])
            
            "Stator Geometry"
            StatorAlpha2[i] = 72/180*np.pi+(Alpha2-72/180*np.pi)/4 if Alpha2 >= 72/180*np.pi else 72/180*np.pi
            StatorInletMachNumber = 1
            StatorDiameter2[i] = Diameter2[i]*(1+(np.pi/2-StatorAlpha2[i])/(2*np.pi)+StatorInletMachNumber**2/15)
            RotorOutletAbsoluteRotationalVelocity[i] = AbsoluteVelocity2*math.sin(Alpha2)
            StatorInletAbsoluteRotationalVelocity[i] = RotorOutletAbsoluteRotationalVelocity[i]*Diameter2[i]/StatorDiameter2[i]
            StatorInletAbsoluteMeridionalVelocity[i] = StatorInletAbsoluteRotationalVelocity[i]/math.tan(StatorAlpha2[i])
            StatorInletAbsoluteVelocity[i] = (StatorInletAbsoluteRotationalVelocity[i]**2+StatorInletAbsoluteMeridionalVelocity[i]**2)**0.5
            StatorInletBladeHeight[i] = Volumeflow2[i]/(StatorInletAbsoluteMeridionalVelocity[i]*np.pi*StatorDiameter2[i])
            StatorVanelessHydraulicLength[i] = (StatorDiameter2[i]-Diameter2[i])/2 
            StatorVanelessHydraulicDiameter[i] = BladeHeight2[i]+StatorInletBladeHeight[i]
            StatorOutletDiameter[i] = Diameter2[i]*(1.55+(TipDiameterRatio[i]**2-HubDiameterRatio[i]**2)*FlowCoefficient)
            StatorOutletBladeHeight[i] = StatorInletBladeHeight[i]
            StatorOutletAbsoluteMeridionalVelocity[i] = Volumeflow3[i]/(np.pi*StatorOutletDiameter[i]*StatorOutletBladeHeight[i])
            StatorOutletAbsoluteVelocity[i] = StatorOutletAbsoluteMeridionalVelocity[i]
            if RotorBladeNumber[i] <= 20:
                StatorBladeNumber[i] = RotorBladeNumber[i]-1
            else:
                StatorBladeNumber[i] = RotorBladeNumber[i]-8
            StatorVanedMeridionalLength[i] = (StatorOutletDiameter[i]-StatorDiameter2[i])/2
            StatorVanedHydraulicLength[i] = (StatorOutletDiameter[i]-StatorDiameter2[i])/(2*math.cos(StatorAlpha2[i]/2))
            StatorInletBladePitch[i] = np.pi*StatorDiameter2[i]/StatorBladeNumber[i]
            StatorOutletBladePitch[i] = np.pi*StatorOutletDiameter[i]/StatorBladeNumber[i]
            StatorSolidity[i] = (StatorOutletDiameter[i]-StatorDiameter2[i])/(2*StatorOutletBladePitch[i]*math.cos(StatorAlpha2[i]/2))
            StatorInletBladeDistance[i] = StatorInletBladePitch[i]*math.cos(StatorAlpha2[i])
            StatorOutletBladeDistance[i] = StatorOutletBladePitch[i]
            StatorInletCrossSection[i] = StatorInletBladeDistance[i]*StatorInletBladeHeight[i]
            StatorOutletCrossSection[i] = StatorOutletBladeDistance[i]*StatorOutletBladeHeight[i]
            StatorInletWettedPerimeter[i] = 2*(StatorInletBladeDistance[i]+StatorInletBladeHeight[i])
            StatorOutletWettedPerimeter[i] = 2*(StatorOutletBladeDistance[i]+StatorOutletBladeHeight[i])
            StatorVanedInletHydraulicDiameter[i] = 2*((StatorInletBladeDistance[i]*StatorInletBladeHeight[i]))/(StatorInletBladeDistance[i]+StatorInletBladeHeight[i])
            StatorVanedOutletHydraulicDiameter[i] = 2*((StatorOutletBladeDistance[i]*StatorOutletBladeHeight[i]))/(StatorOutletBladeDistance[i]+StatorOutletBladeHeight[i])
            StatorVanedHydraulicDiameter[i] = (StatorVanedInletHydraulicDiameter[i]+StatorVanedOutletHydraulicDiameter[i])/2
            
            RotorInletReynoldsNumber[i] = Density1[i]*RelativeVelocity1*RotorHydraulicDiameter[i]/Viscosity1[i]
            RotorOutletReynoldsNumber[i] = Density2[i]*RelativeVelocity2*RotorHydraulicDiameter[i]/Viscosity2[i]
            StatorInletReynoldsNumber[i] = Density2[i]*AbsoluteVelocity2*StatorVanedHydraulicDiameter[i]/Viscosity2[i]
            StatorOutletReynoldsNumber[i] = Density3[i]*AbsoluteVelocity3[i]*StatorVanedHydraulicDiameter[i]/Viscosity3[i]
            
            RotorRoughness[i] = RotorHydraulicDiameter[i]*100/RotorInletReynoldsNumber[i]
            StatorRoughness[i] = StatorVanedHydraulicDiameter[i]*100/StatorInletReynoldsNumber[i]
            
            StatorBladeThickness[i] = 0.01*Diameter2[i]
            
            "_____________________________________________________________________________"
            "Losses"
            "Rotor Losses"
            # Incidence Loss
            IncidenceLossCoefficient[i] = (RotorBladeNumber[i]*RotorBladeThickness[i]/(np.pi*MeanDiameter1[i]*math.cos(Beta1MeanDiameter[i])))**2
            MeanRelativeVelocity1[i] = AbsoluteMeridionalVelocity1/math.cos(Beta1MeanDiameter[i])
            # Skin Friction Loss
            MeanRelativeVelocity[i] = ((MeanRelativeVelocity1[i]**2+RelativeVelocity2**2)/2)**0.5
            RotorSkinFrictionReynoldsNumber[i] = Density1[i]*MeanRelativeVelocity1[i]*RotorHydraulicDiameter[i]/Viscosity1[i]
            if RotorSkinFrictionReynoldsNumber[i] < 2000:
                RotorSkinFrictionCoefficient[i] = 16/RotorSkinFrictionReynoldsNumber[i]
            elif RotorSkinFrictionReynoldsNumber[i] > 4000:
                RotorSkinFrictionCoefficientSmooth[i] = fsolve(SkinFrictionFunctionRotor, 0.001)
                RotorSkinFrictionCoefficientRough[i] = (1/(-4*math.log10(1/3.71*RotorRoughness[i]/RotorHydraulicDiameter[i])))**2
                RotorSkinFrictionRoughReynoldsNumber[i] = (RotorSkinFrictionReynoldsNumber[i]-2000)*RotorRoughness[i]/RotorHydraulicDiameter[i]
                if RotorSkinFrictionRoughReynoldsNumber[i] < 60:
                    RotorSkinFrictionCoefficient[i] = RotorSkinFrictionCoefficientSmooth[i]
                else:
                    RotorSkinFrictionCoefficient[i] = RotorSkinFrictionCoefficientSmooth[i]+(RotorSkinFrictionCoefficientRough[i]-RotorSkinFrictionCoefficientSmooth[i])*(1-60/RotorSkinFrictionRoughReynoldsNumber[i])
            else:
                RotorSkinFrictionRoughReynoldsNumber[i] = (4000-2000)*RotorRoughness[i]/RotorHydraulicDiameter[i]
                if RotorSkinFrictionRoughReynoldsNumber[i] < 60:
                    RotorSkinFrictionCoefficient4000 = fsolve(SkinFrictionFunctionRotor4000, 0.001)
                else:
                    RotorSkinFrictionCoefficientSmooth4000 = fsolve(SkinFrictionFunctionRotor4000, 0.001)
                    RotorSkinFrictionCoefficientRough4000 = (1/(-4*math.log10(1/3.71*RotorRoughness/RotorHydraulicDiameter)))**2
                    RotorSkinFrictionCoefficient4000 = RotorSkinFrictionCoefficientSmooth4000[i]+(RotorSkinFrictionCoefficientRough4000[i]-RotorSkinFrictionCoefficientSmooth4000[i])*(1-60/RotorSkinFrictionRoughReynoldsNumber[i])
                RotorSkinFrictionCoefficient[i] = 16/2000-(16/2000-RotorSkinFrictionCoefficient4000)*(RotorSkinFrictionReynoldsNumber/2000-1)
            RotorSkinFrictionLossCoefficient[i] = 4*RotorSkinFrictionCoefficient[i]*RotorHydraulicLength[i]/RotorHydraulicDiameter[i]*(MeanRelativeVelocity[i]/MeanRelativeVelocity1[i])**2
            # Blade Loading Loss
            BladeLoadingLossCoefficient[i] = 1/24*(2*np.pi*Diameter2[i]*RotationalVelocity2*WorkCoefficient/(RotorBladeNumber[i]*RotorHydraulicLength[i]*MeanRelativeVelocity1[i]))**2
            # Hub to Shroud Losses
            Km[i] = np.pi/2*RotorHydraulicLength[i]
            MeanBladeHeight[i] = (BladeHeight1[i]+BladeHeight2[i])/2
            MeanRelativeVelocityAlt[i] = (MeanRelativeVelocity1[i]+RelativeVelocity2)/2
            HubToShroudLossCoefficient[i] = 1/6*(Km[i]*MeanBladeHeight[i]*MeanRelativeVelocityAlt[i]/MeanRelativeVelocity1[i])**2
            # Mixing Losses
            EquivalentDiffusionFactor[i] = (MeanRelativeVelocity1[i]+RelativeVelocity2+2*np.pi*Diameter2[i]*RotationalVelocity2*WorkCoefficient/(RotorBladeNumber[i]*RotorHydraulicLength[i]))/(2*RelativeVelocity2)
            if EquivalentDiffusionFactor[i] <= 2:
                SeperationVelocity[i] = RelativeVelocity2
            else:
                SeperationVelocity[i] = RelativeVelocity2*EquivalentDiffusionFactor[i]/2
            AbsoluteMeridionalMixingVelocity[i] = AbsoluteMeridionalVelocity2*(1-RotorBladeNumber[i]*RotorBladeThickness[i]/(np.pi*Diameter2[i]))
            AbsoluteMeridionalWakeVelocity[i] = (SeperationVelocity[i]**2-RelativeRotationalVelocity2**2)**0.5
            MixingLossCoefficient[i] = ((AbsoluteMeridionalWakeVelocity[i]-AbsoluteMeridionalMixingVelocity[i])/MeanRelativeVelocity1[i])**2
            # Clearence Losses (only for open impellers, will be implemented at a later time)
            ClearenceLossCoefficient[i] = 0
            
            RotorLossCoefficient[i] = (IncidenceLossCoefficient[i]+RotorSkinFrictionLossCoefficient[i]+BladeLoadingLossCoefficient[i]+HubToShroudLossCoefficient[i]+MixingLossCoefficient[i]+ClearenceLossCoefficient[i])
            
            "Stator Losses"
            # Vaneless Diffuser
            AbsoluteQuadraticMeanVelocity[i] = ((StatorInletAbsoluteVelocity[i]**2+AbsoluteVelocity2**2)/2)**0.5
            DivergenceParameter[i] = BladeHeight2[i]*(StatorDiameter2[i]/Diameter2[i]-1)/StatorVanelessHydraulicLength[i]
            ReferenceParameter[i] = 0.4*(BladeHeight2[i]/StatorVanelessHydraulicLength[i])**0.35
            if DivergenceParameter[i] <= 0:
                DiffusionEfficiency[i] = 1
            elif DivergenceParameter[i] < ReferenceParameter[i]:
                DiffusionEfficiency[i] = 1-0.2*(DivergenceParameter[i]/ReferenceParameter[i])**2
            else:
                DiffusionEfficiency[i] = 0.8*(ReferenceParameter[i]/DivergenceParameter[i])**0.5
            StatorVanelessDiffusionLossCoefficient[i] = -2*(1-DiffusionEfficiency[i])*(StatorInletAbsoluteVelocity[i]-AbsoluteVelocity2)/AbsoluteVelocity2
            
            StatorVanelessSkinFrictionReynoldsNumber[i] = Density2[i]*RelativeVelocity2*StatorVanelessHydraulicDiameter[i]/Viscosity2[i]
            if StatorVanelessSkinFrictionReynoldsNumber[i] < 2000:
                StatorVanelessSkinFrictionCoefficient[i] = 16/StatorVanelessSkinFrictionReynoldsNumber[i]
            elif StatorVanelessSkinFrictionReynoldsNumber[i] > 4000:
                StatorVanelessSkinFrictionCoefficientSmooth[i] = fsolve(SkinFrictionFunctionStatorVaneless, 0.001)
                StatorVanelessSkinFrictionCoefficientRough[i] = (1/(-4*math.log10(1/3.71*StatorRoughness[i]/StatorVanelessHydraulicDiameter[i])))**2
                StatorVanelessSkinFrictionRoughReynoldsNumber[i] = (StatorVanelessSkinFrictionReynoldsNumber[i]-2000)*StatorRoughness[i]/StatorVanelessHydraulicDiameter[i]
                if StatorVanelessSkinFrictionRoughReynoldsNumber[i] < 60:
                    StatorVanelessSkinFrictionCoefficient[i] = StatorVanelessSkinFrictionCoefficientSmooth[i]
                else:
                    StatorVanelessSkinFrictionCoefficient[i] = StatorVanelessSkinFrictionCoefficientSmooth[i]+(StatorVanelessSkinFrictionCoefficientRough[i]-StatorVanelessSkinFrictionCoefficientSmooth[i])*(1-60/StatorVanelessSkinFrictionRoughReynoldsNumber[i])
            else:
                StatorVanelessSkinFrictionRoughReynoldsNumber[i] = (4000-2000)*StatorRoughness[i]/StatorVanelessHydraulicDiameter[i]
                if StatorVanelessSkinFrictionRoughReynoldsNumber[i] < 60:
                    StatorVanelessSkinFrictionCoefficient4000 = fsolve(SkinFrictionFunctionStatorVaneless4000, 0.001)
                else:
                    StatorVanelessSkinFrictionCoefficientSmooth4000 = fsolve(SkinFrictionFunctionStatorVaneless4000, 0.001)
                    StatorVanelessSkinFrictionCoefficientRough4000 = (1/(-4*math.log10(1/3.71*StatorRoughness[i]/StatorVanelessHydraulicDiameter[i])))**2
                    StatorVanelessSkinFrictionCoefficient4000 = StatorVanelessSkinFrictionCoefficientSmooth4000[i]+(StatorVanelessSkinFrictionCoefficientRough4000[i]-StatorVanelessSkinFrictionCoefficientSmooth4000[i])*(1-60/StatorVanelessSkinFrictionRoughReynoldsNumber[i])
                StatorVanelessSkinFrictionCoefficient[i] = 16/2000-(16/2000-StatorVanelessSkinFrictionCoefficient4000)*(StatorVanelessSkinFrictionReynoldsNumber/2000-1)
            StatorVanelessSkinFrictionLossCoefficient[i] = 4*StatorVanelessSkinFrictionCoefficient[i]*StatorVanelessHydraulicLength[i]/StatorVanelessHydraulicDiameter[i]*(MeanRelativeVelocity[i]/MeanRelativeVelocity1[i])**2
            
            StatorVanelessLossCoefficient[i] = (StatorVanelessDiffusionLossCoefficient[i]+StatorVanelessSkinFrictionLossCoefficient[i])
            
            # Vaned Diffuser
            StatorVanedIncidenceLossCoefficient[i] = 0.8*(1-StatorInletAbsoluteMeridionalVelocity[i]/(StatorInletAbsoluteVelocity[i]*math.cos(StatorAlpha2[i])))**2+(StatorBladeNumber[i]*StatorBladeThickness[i]/(np.pi*StatorDiameter2[i]))**2
            AbsoluteMeridionalMixingVelocity3[i] = StatorOutletAbsoluteMeridionalVelocity[i]*(1-StatorBladeNumber[i]*StatorBladeThickness[i]/(np.pi*StatorOutletDiameter[i]))
            StatorEquivalentDiffusionFactor[i] = StatorInletAbsoluteVelocity[i]/StatorOutletAbsoluteVelocity[i]
            if StatorEquivalentDiffusionFactor[i] <= 2:
                SeperationVelocity3[i] = StatorOutletAbsoluteVelocity[i]
            else:
                SeperationVelocity3[i] = StatorOutletAbsoluteVelocity[i]*StatorEquivalentDiffusionFactor[i]/2
            AbsoluteMeridionalWakeVelocity3[i] = (SeperationVelocity3[i]**2-StatorOutletAbsoluteVelocity[i]**2)**0.5
            StatorVanedMixingLossCoefficient[i] = ((AbsoluteMeridionalWakeVelocity3[i]-AbsoluteMeridionalMixingVelocity3[i])/StatorInletAbsoluteVelocity[i])**2
            
            StatorVanedSkinFrictionReynoldsNumber[i] = Density2[i]*StatorInletAbsoluteVelocity[i]*StatorVanedHydraulicDiameter[i]/Viscosity2[i]
            if StatorVanedSkinFrictionReynoldsNumber[i] < 2000:
                StatorVanedSkinFrictionCoefficient[i] = 16/StatorVanedSkinFrictionReynoldsNumber[i]
            elif StatorVanedSkinFrictionReynoldsNumber[i] > 4000:
                StatorVanedSkinFrictionCoefficientSmooth[i] = fsolve(SkinFrictionFunctionStatorVaned, 0.001)
                StatorVanedSkinFrictionCoefficientRough[i] = (1/(-4*math.log10(1/3.71*StatorRoughness[i]/StatorVanedHydraulicDiameter[i])))**2
                StatorVanedSkinFrictionRoughReynoldsNumber[i] = (StatorVanedSkinFrictionReynoldsNumber[i]-2000)*StatorRoughness[i]/StatorVanedHydraulicDiameter[i]
                if StatorVanedSkinFrictionRoughReynoldsNumber[i] < 60:
                    StatorVanedSkinFrictionCoefficient[i] = StatorVanedSkinFrictionCoefficientSmooth[i]
                else:
                    StatorVanedSkinFrictionCoefficient[i] = StatorVanedSkinFrictionCoefficientSmooth[i]+(StatorVanedSkinFrictionCoefficientRough[i]-StatorVanedSkinFrictionCoefficientSmooth[i])*(1-60/StatorVanedSkinFrictionRoughReynoldsNumber[i])
            else:
                StatorVanedSkinFrictionRoughReynoldsNumber[i] = (4000-2000)*StatorRoughness[i]/StatorVanedHydraulicDiameter
                if StatorVanedSkinFrictionRoughReynoldsNumber[i] < 60:
                    StatorVanedSkinFrictionCoefficient4000 = fsolve(SkinFrictionFunctionStatorVaned4000, 0.001)
                else:
                    StatorVanedSkinFrictionCoefficientSmooth4000 = fsolve(SkinFrictionFunctionStatorVaned4000, 0.001)
                    StatorVanedSkinFrictionCoefficientRough4000 = (1/(-4*math.log10(1/3.71*StatorRoughness/StatorVanedHydraulicDiameter)))**2
                    StatorVanedSkinFrictionCoefficient4000 = StatorVanedSkinFrictionCoefficientSmooth4000[i]+(StatorVanedSkinFrictionCoefficientRough4000[i]-StatorVanedSkinFrictionCoefficientSmooth4000[i])*(1-60/StatorVanedSkinFrictionRoughReynoldsNumber[i])
                StatorVanedSkinFrictionCoefficient[i] = 16/2000-(16/2000-StatorVanedSkinFrictionCoefficient4000)*(StatorVanedSkinFrictionReynoldsNumber/2000-1)
            A[i] = (5.142*StatorVanedSkinFrictionCoefficient[i]*StatorVanedHydraulicLength[i]/StatorVanedHydraulicDiameter[i])**0.25
            StatorVanedSkinFrictionLossCoefficient[i] = 4*StatorVanedSkinFrictionCoefficient[i]/A[i]*StatorVanedHydraulicLength[i]/StatorVanedHydraulicDiameter[i]*(AbsoluteQuadraticMeanVelocity[i]/StatorInletAbsoluteVelocity[i])**2
            
            StatorVanedLossCoefficient[i] = (StatorVanedIncidenceLossCoefficient[i]+StatorVanedMixingLossCoefficient[i]+StatorVanedSkinFrictionLossCoefficient[i])
            
            StatorLossCoefficient[i] = (StatorVanelessLossCoefficient[i]+StatorVanedLossCoefficient[i])
            
            
            "_____________________________________________________________________________"
            "Efficiency Calculations"
            IsentropicTotalRelativeEnthalpy2[i] = TotalRelativeEnthalpy2[i]  # IsentropicEnthalpy2[i]+RelativeVelocity2**2/2 
            IsentropicTotalRelativePressure2[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, IsentropicTotalRelativeEnthalpy2[i], Entropy1[i], WorkingFluidComposition).Output[0]
            LossTotalRelativePressure2[i] = IsentropicTotalRelativePressure2[i]/(1+RotorLossCoefficient[i]*(1-Pressure1[i]/TotalRelativePressure1[i]))
            LossEntropy2[i] = REFPROP.refprop(WorkingFluid, "HP", "S", Units, 0, 0, TotalRelativeEnthalpy2[i], LossTotalRelativePressure2[i], WorkingFluidComposition).Output[0]
            LossEnthalpy2[i] = REFPROP.refprop(WorkingFluid, "SP", "H", Units, 0, 0, LossEntropy2[i], Pressure2[i], WorkingFluidComposition).Output[0]
            RotorEfficiency[i] = (IsentropicEnthalpy2[i]-Enthalpy1[i])/(LossEnthalpy2[i]-Enthalpy1[i])
            LossTotalEnthalpy2[i] = LossEnthalpy2[i]+AbsoluteVelocity2**2/2 
            LossTotalPressure2[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, LossTotalEnthalpy2[i], LossEntropy2[i], WorkingFluidComposition).Output[0]
            LossPressure2[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, LossEnthalpy2[i], LossEntropy2[i], WorkingFluidComposition).Output[0]
            LossStatorInletTotalPressure2[i] = LossTotalPressure2[i]-StatorVanelessLossCoefficient[i]*(LossTotalPressure2[i]-LossPressure2[i])
            LossTotalPressure3[i] = LossStatorInletTotalPressure2[i]-StatorVanedLossCoefficient[i]*(LossStatorInletTotalPressure2[i]-LossPressure2[i])
            LossTotalEnthalpy3[i] = LossTotalEnthalpy2[i]
            LossEntropy3[i] = REFPROP.refprop(WorkingFluid, "HP", "S", Units, 0, 0, LossTotalEnthalpy3[i], LossTotalPressure3[i], WorkingFluidComposition).Output[0]
            LossEnthalpy3[i] = REFPROP.refprop(WorkingFluid, "SP", "H", Units, 0, 0, LossEntropy3[i], Pressure3[i], WorkingFluidComposition).Output[0]
            LossTotalEnthalpy3[i] = LossEnthalpy3[i]+AbsoluteVelocity3[i]**2/2 
            TotalEfficiency[i] = ((IsentropicEnthalpy3[i]+AbsoluteVelocity3[i]**2/2)-TotalEnthalpy1[i])/(LossTotalEnthalpy3[i]-TotalEnthalpy1[i])
            IsentropicEfficiency[i] = (IsentropicEnthalpy3[i]-Enthalpy1[i])/(LossEnthalpy3[i]-Enthalpy1[i])
            StaticEfficiency[i] = (IsentropicEnthalpy3[i]-TotalEnthalpy1[i])/(LossTotalEnthalpy3[i]-TotalEnthalpy1[i])
            StatorEfficiency[i] = (IsentropicEnthalpy3[i]-LossEnthalpy2[i])/(LossEnthalpy3[i]-LossEnthalpy2[i])
        
            "Thermodynamic Inlet Conditions of the following Stage"
            if i <StageNumber -1:
                Enthalpy1[i+1] = LossEnthalpy3[i]+AbsoluteVelocity3[i]**2/2-AbsoluteVelocity1**2/2 
                Entropy1[i+1] = LossEntropy3[i]
                Pressure1[i+1] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, LossEnthalpy3[i], LossEntropy3[i], WorkingFluidComposition).Output[0]
                Temperature1[i+1] = REFPROP.refprop(WorkingFluid, "HS", "T", Units, 0, 0, LossEnthalpy3[i], LossEntropy3[i], WorkingFluidComposition).Output[0]
                Density1[i+1] = REFPROP.refprop(WorkingFluid, "HS", "D", Units, 0, 0, LossEnthalpy3[i], LossEntropy3[i], WorkingFluidComposition).Output[0]
                
            "Loss based Properties"
            LossTemperature2[i] = REFPROP.refprop(WorkingFluid, "HS", "T", Units, 0, 0, LossEnthalpy2[i], LossEntropy2[i], WorkingFluidComposition).Output[0]
            LossTemperature3[i] = REFPROP.refprop(WorkingFluid, "HP", "T", Units, 0, 0, LossEnthalpy3[i], Pressure3[i], WorkingFluidComposition).Output[0]
            LossPressure2[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, LossEnthalpy2[i], LossEntropy2[i], WorkingFluidComposition).Output[0]
            LossPressure3[i] = REFPROP.refprop(WorkingFluid, "HS", "P", Units, 0, 0, LossEnthalpy3[i], LossEntropy3[i], WorkingFluidComposition).Output[0]
            
            # End of the loop over the stages
        EstimatedIsentropicRotorEfficiency= RotorEfficiency
        OverallEstimatedIsentropicEfficiency = (IsentropicOutletEnthalpy-InletEnthalpy)/((LossEnthalpy3[StageNumber-1]-InletEnthalpy))
        # if OverallEstimatedIsentropicEfficiency <= 0:       # Check to ensure that value is physically feasible for next iteration step
        #     OverallEstimatedIsentropicEfficiency = 1
        # print("Estimation")
        # print(OverallEstimatedIsentropicEfficiency)
        CalculatedIsentropicEfficiency[j] = OverallEstimatedIsentropicEfficiency
        
        "Losses and Efficiency"
        # fig, axs = plt.subplots(1, 2, figsize=(15, 10))
        # fig.suptitle('Losses and Efficiency')
        # x = np.array(range(StageNumber))+1
        # axs[0].plot(x, RotorLossCoefficient, 'x-', label='Rotor')
        # axs[0].plot(x, StatorLossCoefficient, 'x-', label='Stator')
        # axs[0].set_title('Loss Coefficients')
        # axs[0].set_xlabel('Stage Number')
        # axs[0].set_ylabel('Loss Coefficients')
        # axs[0].legend()
    
        # axs[1].plot(x, IsentropicEfficiency, 'x-', label='Isentropic')
        # axs[1].plot(x, StaticEfficiency, 'x-', label='Static')
        # axs[1].plot(x, TotalEfficiency, 'x-', label='Total')
        # axs[1].plot(x, RotorEfficiency, 'x-', label='Rotor')
        # axs[1].set_title('Efficiencies')
        # axs[1].set_xlabel('Stage Number')
        # axs[1].set_ylabel('Efficiency')
        # axs[1].legend()
    
        # plt.tight_layout(rect=[0, 0.03, 1, 0.95])
        # plt.show()
        # End of the loop redoing the calculations
        
    "_____________________________________________________________________________"
    "Machine Results"
    OverallIsentropicEfficiency = (IsentropicOutletEnthalpy-InletEnthalpy)/((LossEnthalpy3[StageNumber-1]-InletEnthalpy))
    OverallTotalEfficiency = (IsentropicOutletEnthalpy+AbsoluteVelocity3**2/2-InletEnthalpy-AbsoluteVelocity1**2/2)/((LossTotalEnthalpy3[StageNumber-1]-InletEnthalpy-AbsoluteVelocity1**2/2))
    OverallStaticEfficiency = (IsentropicOutletEnthalpy-InletEnthalpy-AbsoluteVelocity1**2/2)/((LossTotalEnthalpy3[StageNumber-1]-InletEnthalpy-AbsoluteVelocity1**2/2))
    # print("Isentropic Efficiency")
    # print(OverallIsentropicEfficiency)
    
    ResultDictionary = {"Isentropic Efficiency": OverallIsentropicEfficiency,
                        "Total Efficiency": OverallTotalEfficiency,
                        "Static Efficiency": OverallStaticEfficiency,
                        "Inlet Tip Diameter": TipDiameter1,
                        "Inlet Hub Diameter": HubDiameter1,
                        "Rotor Outlet Diameter": Diameter2,
                        "Stator Inlet Diameter": StatorDiameter2,
                        "Stator Outlet Diameter": StatorOutletDiameter,
                        "Rotor Blade Number": RotorBladeNumber,
                        "Stator Blade Number": StatorBladeNumber,
                        "Rotor Axial Lenght": RotorAxialLength,
                        "Inlet Velocity": AbsoluteVelocity1,
                        "Outlet Velocity": AbsoluteVelocity3[StageNumber-1],
                        "Outlet Temperature": Temperature3[StageNumber-1],
                        "Outlet Pressure": Pressure3[StageNumber-1],
                        "Outlet Enthalpy": Enthalpy3[StageNumber-1],
                        "Outlet Entropy": Entropy3[StageNumber-1],
                        "Outlet Density": Density3[StageNumber-1],
                        "Stage Number": StageNumber,
                        "Work Coefficient": WorkCoefficient,
                        "Flow Coefficient": FlowCoefficient,
                        "Specific Speed": SpecificSpeed,
                        "Specific Diameter": SpecificDiameter}


    return ResultDictionary


# -*- coding: utf-8 -*-
"""
Scripts for the evaluation of the results of the fluid search, e.g. finding

pareto optimal compositions.

Created on Sun Feb  4 17:38:19 2024

@author: atakan
part of carbatpy
"""
import os
import pandas

from paretoset import paretoset
import seaborn as sbn


def pareto(filename, objectives):
    """
    Compute the Pareto (non-dominated) set from the data stored in filename



    Parameters
    ----------
    filename : string
        directory/file of the csv-file to analyze. The column names will be
        used to find the objectives.
    objectives : list of strings, length [2, number_of_objectives]
        the first list in the list are the column names to search for optima.
        the second list indicates for each optimization, whether a "max" or
        "min" is searched. Here "diff" can be in a column to indicate a column
        with categories, which are analyzed separately.

    Returns
    -------
     results: dictionary
         with the whole pandas dataFrame (key:"all_values") and a list of
         booleans if True, this line of the set is pareto optimal
         (key:"optimal_mask"). Also, the "objectives" are returned.

    """
    results = {}
    try:
        results["all_values"] = pandas.read_csv(filename)
    except Exception as excep:
        text = f"{filename} vs. {os.getcwd()}, {excep}"
        raise ValueError(text) from excep

    c_names = results["all_values"].columns
    if set(objectives[0]).issubset(set(c_names)):

        results["optimal_mask"] = paretoset(
            results["all_values"][objectives[0]], sense=objectives[1])
        results["objectives"] = objectives
        return results

    print("keys do not fit, they should be one of:", c_names)
    return []


if __name__ == "__main__":
    FILENAME_ACT = r"C:\Users\atakan\sciebo\Python\carbatpy\tests\test_files\test_data_ProEthPenBut"
    FILENAME_ACT += r"\\2024-02-06-16-51-ProEthPenBut"
    objectives_act = ['p_ratio', 'T_glide_h', 'spec_Volume_sup', 'COP_is80'] # 'spec. Volume_sup',
    sense = ["min", "min", "min", "max"]
    obj_sense = [objectives_act, sense]
    res = pandas.read_csv(FILENAME_ACT+".csv")

    opti = pareto(FILENAME_ACT+".csv", obj_sense)
    optimal_data = opti["all_values"][opti["optimal_mask"]]
    optimal_data.to_csv(FILENAME_ACT+"-pareto.csv")
    with open(FILENAME_ACT+"-objectives.txt", "w", encoding="utf-8") as file:
              file.write(str(list(zip(objectives_act,sense))))

    graph = sbn.relplot(data=optimal_data,
                        y=objectives_act[-1],
                        hue=objectives_act[0],
                        size=objectives_act[1],
                        x=r'spec_Volume_sup')
    graph.savefig(FILENAME_ACT+"-plot.jpg", dpi =300)



# -*- coding: utf-8 -*-
"""
Created on Sun May 21 10:52:09 2023

@author: atakan
"""

import copy
import numpy as np
import matplotlib.pyplot as plt

import carbatpy as cb

_RESULTS_ = r"C:\Users\atakan\sciebo\results\\"


class HeatPump:
    """
    Heat pump class for simple thermodynamic calculations for energy storage

    Two pairwase storages at two different temperatures are assumed,
    always starting at room temperature (around 300K). Thus, three fluids are
    expected: the working fluid, the storage fluid  or sink at high temperature
    (h) and the source storage fluid at low temperature.
    The minimum and maximum temperatures of all 
    storags are fixed and the heat flow rate to the high-T storage are 
    given. Also, isentropic efficiencies are fixed so far, but they should use
    a function later.
    """

    def __init__(self,  fluids, fixed_points, components=[]):
        """
        Heat pump class for simple thermodynamic calculations for energy storage

        storages at two different temperatures, always starting at room 
        temperature (around 300K). The minimum and maximum temperatures aof all 
        storags are fixed and the heat flow rate to the high-T storage are 
        given. Also, isentropic efficiencies.
        Example for
        fixed_points = {"eta_s": 0.65,
                        "p_low": p_low,
                        "T_hh": _STORAGE_T_OUT_,
                        "T_hl": _STORAGE_T_IN_,
                        "T_lh": _STORAGE_T_IN_,
                        "T_ll": 257.,  # 256.0,
                        "Q_dot_h": 3000.0,
                        "d_temp_min": _D_T_MIN_}

        Parameters
        ----------
        fluids : list of three Fluid
            working fluid, secondary fluid to store at high T, cold fluid
            to store at low T.
        fixed_points : Dictionary
            all the fixed points are summarized here, the isentropic efficiency,
            the lower pressure(evaporator), the high and low temperatures of
            the high temperature storages, the same for the low temperature
            storages(source), the heat to be transfered at high T, the minimum
            approach temperature.
        components : TYPE, optional
            DESCRIPTION. The default is [].

        Returns
        -------
        None.

        """
        self.components = components
        self.fixed_points = fixed_points
        self.fluids = fluids
        self.all_states = []
        self.m_dots = []
        self.evaluation = {"Q_dot_h": self.fixed_points["Q_dot_h"],
                           "Power": 0.0,
                           "p_high": 0.0,
                           "p_low!": self.fixed_points["p_low"],
                           "exergy_loss_rate": 0}

    def calc_p_high(self, p_high, verbose=True):
        """
        calculate a simple compression heat pump 

        working with two sensible storages (or source and sink). At the moment
        the compressor model is just for a constant isentropic efficiency.
        An isenthalpic throttle is used for expnsion. The initial states are provided
        by the Fluid-instances. The pressures have to be selected carefully,
        they are not tested at the moment. The high pressure is an input, and
        can easily be used for optimization. Several parameters are fixed
        along a calculation, this is in the dictionary "self.fixed_points".
        It can turn out that the foreseen evaporator exit temperature of the 
        secondary(cooling) fluid is below the temperature of the isenthalpic 
        temperature of the fluid coming out of the throttle. In this case the 
        secondary exit temperature is increased.

        Parameters
        ----------
        p_high : float
            working fluid pressure (Pa) in the condenser (=const.).
        verbose : Boolean, optional
            should values be printed? The deault is False

        Returns
        -------
        COP : flaot
            coefficient of performace.


        """

        state_in = copy.copy(self.fluids[0])
        w_actual = self.fluids[0]
        if verbose:
            print(
                "Initial:", self.fluids[0].properties.temperature,
                state_in.properties.enthalpy)
        w_after_compressor = \
            cb.compressor_simple.compressor(p_high, self.fixed_points["eta_s"],
                                            self.fluids[0])
        if verbose:
            print("behind compressor",
                  self.fluids[0].properties.temperature, w_after_compressor)

        condenser = \
            cb.hex_th.StaticHeatExchanger([w_actual,
                                           self.fluids[1]],
                                          self.fixed_points["Q_dot_h"],
                                          self.fixed_points["T_hh"],
                                          d_temp_separation_min=self.fixed_points["d_temp_min"],
                                          name="condenser")
        factor = condenser.find_pinch()
        m_dot_wf, d_tempall, wf_states, sf_states =\
            condenser.pinch_plot(factor, plotting=False)
        self.all_states.append(wf_states)
        self.all_states.append(sf_states)
        self.m_dots.append(m_dot_wf)
        self.m_dots.append(condenser.m_dot_s)

        w_after_condenser = self.fluids[0].set_state([wf_states[0, 1],
                                                      wf_states[0, 2]], "PH")
        if verbose:
            print("after condenser")
            w_actual.print_state()

        power = (w_after_compressor[2] -
                 state_in.properties.enthalpy) * m_dot_wf
        heat_flow_evap = self.fixed_points["Q_dot_h"] - power
        if verbose:
            print(
                f"power:{ power} W, heat-Evap. {heat_flow_evap} W\n")
        w_after_throttle = cb.throttle_simple.throttle(self.fixed_points["p_low"],
                                                       self.fluids[0])

        if verbose:
            print("after throttle:")
        w_act = w_actual.set_state(
            [w_after_throttle[1], w_after_throttle[2]], "PH")
        if self.fixed_points["T_ll"] < w_act[0] + self.fixed_points["d_temp_min"]:
            self.fixed_points["T_ll"] = w_act[0] + \
                self.fixed_points["d_temp_min"]
        if verbose:
            w_actual.print_state()

        evaporator = cb.hex_th.StaticHeatExchanger([w_actual, self.fluids[2]],
                                                   heat_flow_evap,
                                                   self.fixed_points["T_ll"],
                                                   d_temp_separation_min=self.fixed_points["d_temp_min"],
                                                   calc_type="const",
                                                   name="evaporator")
        # factor_ev = evaporator.find_pinch()[0]
        # problem, must perhaps be rewritten, because the mass flow in the cycle is fixed:
        mw_ev, d_tempall_ec, w_ev, s_ev = evaporator.pinch_plot(
            factor, plotting=False)

        self.all_states.append(w_ev)
        self.all_states.append(s_ev)
        self.m_dots.append(m_dot_wf)
        self.m_dots.append(evaporator.m_dot_s*m_dot_wf/mw_ev)
        w_after_evap = self.fluids[0].set_state([w_ev[0, 1], w_ev[0, 2]], "PH")
        q_evap = (w_after_evap[2] - w_after_throttle[2]) * m_dot_wf
        if heat_flow_evap - q_evap > 1:
            print(f"heat pump cycle not steady{heat_flow_evap, q_evap}")
        if verbose:
            print(f"  heat after Evaporator:{q_evap},power:{power}")
            w_actual.print_state()
            print("-test-", (sf_states[-1] - sf_states[0])*self.m_dots[1],
                  (s_ev[-1]-s_ev[0])*self.m_dots[-1])
        sec_fluids_states_in_out = np.array([[sf_states[0], sf_states[-1]],
                                            [s_ev[0], s_ev[-1]]])
        
        mass_flow_rates = np.array([self.m_dots[1], self.m_dots[-1]])

        ex_loss_rate = cb.exlo.exergy_loss_flow(sec_fluids_states_in_out,
                                                mass_flow_rates)
        ex_loss_alte = cb.exlo.exergy_loss_alternative(sec_fluids_states_in_out,
                                                       mass_flow_rates,
                                                       power_in=power)
        self.evaluation = {"Q_dot_h": self.fixed_points["Q_dot_h"],
                           "Power": power,
                           "p_high": p_high,
                           "p_low": self.fixed_points["p_low"],
                           "exergy_loss_rate": ex_loss_rate,
                           "exergy_loss_alte": ex_loss_alte}
        return self.fixed_points["Q_dot_h"] / power

    def hp_plot(self, f_name=_RESULTS_+"last_T_H_dot_plot"):
        """
        plots the heat pump cycle and stores it to the given file (name)

        Returns
        -------
        None.

        """
        states = []
        connect_temp = []
        connect_h = []
        position = [0, -1]
        fig, ax = plt.subplots(1, 1)
        points = ["-k", "or", "-b", "vg"]
        m_dots = self.m_dots

        m_dots[3] = m_dots[3] / m_dots[2] * m_dots[0]
        m_dots[2] = m_dots[0]

        for i_process in range(4):

            cycle_states = self.all_states[i_process]
            cycle_states = cycle_states[cycle_states[:, 2].argsort(), :]
            if i_process == 2:
                liquid = self.fluids[0].set_state([cycle_states[0, 1], 0.],
                                                  "PQ")
                ax.plot((liquid[2] - cycle_states[0, 2]) * m_dots[i_process],
                        liquid[0], "o")
            cycle_states[:, 2] -= cycle_states[0, 2]

            cycle_states[:, 2] = cycle_states[:, 2] * m_dots[i_process]
            states.append(cycle_states)

            ax.plot(cycle_states[:, 2], cycle_states[:, 0], points[i_process])
            if i_process == 0 or i_process == 2:
                connect_temp.append(cycle_states[position, 0])
                connect_h.append(cycle_states[position, 2])
        temperatures = np.array(connect_temp).reshape((4))
        enthalpies = np.array(connect_h).reshape((4))
        ax.set_xlabel("$\dot H$ / W")
        ax.set_ylabel("$T$ / K")

        for i_process in range(2):
            ax.plot([enthalpies[0+i_process], enthalpies[2+i_process]],
                    [temperatures[0+i_process], temperatures[2+i_process]], "r:")
        fig.savefig(f_name+".png")


if __name__ == "__main__":

    FLUID = "Propane * Butane * Pentane * Hexane"
    comp = [.75, 0.05, 0.15, 0.05]
    comp = [0.4,	0.3,	0.3, 0.0]  # [0.164,.3330,.50300,0.0]

    FLS = "Water"  #
    FLCOLD = "Methanol"  # "Water"  #

    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])

    coldFlm = cb.fprop.FluidModel(FLCOLD)
    coldFluid = cb.fprop.Fluid(coldFlm, [1.])

    # Condenser(c) and storage (s), secondary fluids fix all, temperatures(T in K),
    # pressures (p in Pa)
    _ETA_S_ = 0.637  # interesting when changed from 0.69 to 0.65, the efficiency
    # decreases, the reason is the low quality along throtteling then
    _STORAGE_T_IN_ = cb._T_SURROUNDING
    _COLD_STORAGE_T_IN_ = _STORAGE_T_IN_
    _STORAGE_T_OUT_ = 363.  # 395.0
    _COLD_STORAGE_T_OUT_ = 260.15-5
    _STORAGE_P_IN_ = 5e5
    _COLD_STORAGE_P_IN_ = 5e5
    _Q_DOT_MIN_ = 1e3  # and heat_flow rate (W)
    _D_T_SUPER_ = 5  # super heating of working fluid
    _D_T_MIN_ = 5.  # minimum approach temperature (pinch point)

    state_sec_in = secFluid.set_state([_STORAGE_T_IN_, _STORAGE_P_IN_], "TP")
    state_cold_in = coldFluid.set_state(
        [_COLD_STORAGE_T_IN_, _COLD_STORAGE_P_IN_], "TP")

    # working fluid
    T_DEW = _STORAGE_T_OUT_ - _D_T_SUPER_
    state_in = myFluid.set_state([T_DEW, 1.], "TQ")  # find minimum pressure
    # changed by hand, to make the _D_T_MIN_ possible
    p_high_act = state_in[1] * .99

    state_in = myFluid.set_state(
        [_STORAGE_T_IN_-_D_T_SUPER_-_D_T_MIN_, 1.], "TQ")  # find minimum pressure
    p_low = state_in[1]
    T_IN = _STORAGE_T_IN_  # - _D_T_MIN_
    # state_in = myFluid.set_state([p_low,
    #                               _STORAGE_T_IN_ - _D_T_MIN_], "PT")
    print(f"p-ratio: {p_high_act/p_low: .2f}, p_low: {p_low/1e5: .2} bar")
    state_in = myFluid.set_state([p_low,
                                  T_IN], "PT")

    FIXED_POINTS = {"eta_s": _ETA_S_,
                    "p_low": p_low,
                    "T_hh": _STORAGE_T_OUT_,
                    "T_hl": _STORAGE_T_IN_,
                    "T_lh": _STORAGE_T_IN_,
                    "T_ll": _COLD_STORAGE_T_OUT_,  # 256.0,
                    "Q_dot_h": _Q_DOT_MIN_,
                    "d_temp_min": _D_T_MIN_}

    hp0 = HeatPump([myFluid, secFluid, coldFluid], FIXED_POINTS)
    cop = hp0.calc_p_high(p_high_act, verbose=True)
    hp0.hp_plot()
    out = hp0.evaluation
    print(
        f"COP: {cop},p-ratio: {out['p_high']/out['p_low']:.2f}, p_low {out['p_low']/1e5:.2f} bar")
    print(
        f"COP: {cop},p-ratio: {out['p_high']/out['p_low']}, p_low {out['p_low']/1e5}")
    print(
        f'exergy loss rate: {out["exergy_loss_rate"]}, eff: {1-out["exergy_loss_rate"]/out["Power"]:.4f}')
    print(
        f'power= {out["Power"]:.2f} W, Alternative Ex-Loss:{out["exergy_loss_alte"]:.4f}')
    print(f'eff_alternative: {1-out["exergy_loss_alte"]/out["Power"]:.4f}')



# -*- coding: utf-8 -*-
"""

A script to evaluate mixtures, in order to find some with vapor pressures
in some limits at given temperatures, together with the temperature glide.
The results are stored as figure, as csv and in a json-file(Input); all in the
given directory.


The csv output file structure is as follows:

* number of calculation
* the four mole fractions, species names are in the title
* index l: the properties for saturated vapor at the given low temperature
* index sup: the poperties at superheating at pressure p_l for a prescribed superheating
* index h: the properties for saturated vapor at the given high temperature
* index is: the properties for the isentropic state (sup ->p_h) at the given low temperature
* index is80: the properties for the isentropic effic. of 80 % (sup ->p_h) at the given low temperature
* index dew: the properties for the saturated liquid at p_h
* index mid: the properties at the mean enthalpy between q=0 and q=1 at p_h
* index thr: the properties for the isenthalpic throtteling from saturated liquid to p_l
* index hplT: the properties at T_l and p_h
* index thrlow: the properties for the isenthalpic throtteling from hplt ->p_l
* index bol: the properties for saturated liquid at the low pressure p_l
* p_ratio: the pressure ratio
* T_glide_h: the temperature glide at high pressure
* dv/v'': (ca.) the mean change in volume along throtteling relative to the specific volume of the vapor, this is a measure of how much work is 'lost' along throtteling
* dv/v''-b: similar volume ratio after subcooling to thrlow, answer the question: will subcooling reduce losses (strongly)?
* COP_is: What is the predicted COP for isentropic compression (losses along throtteling are seen here)

For each indexed state : T,p,h,v,s,q,u in SI units(mass base) are listed.


part of carbatpy

Created on Thu Oct 19 14:11:14 2023

@author: atakan
"""

import os
import json
import itertools
import matplotlib.pyplot as plt
import seaborn as sbn
import pandas as pd
import numpy as np
import carbatpy as cb


def mixture_search(fluids_all, temp_both, p_both, res_dir, d_temp_superheating=5,
                   resolution=21, temp_limit=False, **kwargs):
    """
    Mixtures are evaluated/screened to find mixtures with a given temperature
    glide in a certain pressure range. For all possible mixture compositions
    first the saturated vapor pressure at the given low temperature and composition
    is evaluated, if it is in the allowed regime, the saturated vapor pressure
    (p_h) of the mixture at the high temperature is evaluated.
    If this is below the allowed high
    pressure, the saturated liquid temperature at this p_h is evaluated and the
    temperature difference is taken as temperature glide. The pressure ratio is
    plotted as a function of temperature glide.
    The plot, the states (as csv-file) and the input parameters (as .json-file)
    are stored.
    in the csv-File first the mole fractions are given followed by the
    properties of the low temperature sat.vapor, after superheating,
    for thehigh temperature sat. vapor, the isentropic state after compression,
    high pressure sat.liquid, low pressuer isenthalpic (throtteling) state to
    the high pressure sat. liquid, and the low pressure saturated liquis state.
    For each of them: T,p,h,v,s,q,u in SI units(mass base)

    Parameters
    ----------
    fluids_all : list of strings
        up to 4 fluid names of the mixture, as defined in the fluid model
        (REFPROP).
    temp_both : List of two floats
        the minimum (at low pressure) and the maximum (at high pressure)
        saturated vapor temperature (both dew points) in K.
    p_both : List of two floats
        allowed min and max pressure in Pa.
    res_dir : string
        Directory name, where the results are stored.
    d_temp_superheating : float. optional
        super heating temperature in K. The default is 5.
    resolution : integer, optional
        inverse is the interval for the mole fraction screening (21 means every
        0.05 a value is calculated). The default is 21.
    temp_limit : Boolean, optional
        selects only values, where the temperature of the saturated liquid at
        high pressure is above temp_both[0]. The default is False.
    kwargs : dict
        is not implemented yet, but can be used later to select another fluid
        model, instead of REFPROP.

    Returns
    -------
    None.

    """
    # dir_name = r"C:\Users\atakan\sciebo\results\optimal_hp_fluid"
    eff_isentropic = 0.8  # isentropic efficiency, compressor
    all_results = {}
    exception_messages = []
    all_results["warn"] = 0
    if len(kwargs) > 0:
        print(f"These arguments are not implemented yet{kwargs}")
        all_results["warn"] = 1
    dir_name = res_dir

    plt.style.use('seaborn-v0_8-poster')  # 'seaborn-v0_8-poster')
    # fluids_all = ["Ethane", "Propane", "Pentane", "CO2"]
    fluid_mixture = "*".join(fluids_all)
    # fluid_mixture = "Dimethylether * Butane * Pentane * Hexane" #  "Propane * Butane * Pentane * Hexane"
    names = ["x_" + s for s in fluids_all]

    fn_end = "".join([s[:3] for s in fluids_all])
    fname = cb.helpers.file_copy.copy_script_add_date(
        fn_end, __file__, dir_name)
    # fname = date + fn_end
    comp = [.5, 0., 0.5, 0.0]  # , 0.0]

    flm = cb.fprop.FluidModel(fluid_mixture)
    my_fluid = cb.fprop.Fluid(flm, comp)
    temp_low, temp_high = temp_both
    p_low, p_high = p_both
    x_i_range = np.linspace(0, 1, resolution)
    results = []
    sound_speeds = []
    lin_t_deviations_all =[]

    n_species = len(fluids_all)

    variables_dict = {"File_name": fname,
                      "Dir_name": dir_name,
                      "Fluid": fluid_mixture,
                      "T_low": temp_low,
                      "T_high": temp_high,
                      "d_temp_superheating": d_temp_superheating,
                      "p_low": p_low,
                      "p_high": p_high,
                      "eta_is80": eff_isentropic,
                      "what": "T_sat_l, T_sat_h, T_super, T_is,T_is80, T_throt,Tsat_l_high"
                      }

    # Dateipfad, in dem die JSON-Datei gespeichert wird
    json_file_path = fname+"_variablen.json"

    # Speichern des Dictionaries in einer JSON-Datei
    with open(json_file_path, 'w', encoding="utf-8") as json_file:
        json.dump(variables_dict, json_file)

    print(f"Variablen wurden in '{json_file_path}' gespeichert.")
    mole_fractions = np.zeros((n_species))

    for positions in itertools.product(range(len(x_i_range)), repeat=n_species-1):
        # positions ist ein Tupel, das die ausgewählten Positionen enthält
        actual_x = np.array([x_i_range[i] for i in positions])
        if actual_x.sum() <= 1:
            mole_fractions[:3] = actual_x
            mole_fractions[n_species-1] = 1 - actual_x.sum()
            my_fluid.set_composition(mole_fractions)

            try:
                state_low = my_fluid.set_state(
                    [temp_low - d_temp_superheating, 1.], "TQ")  # find low pressure
                state_sup = my_fluid.set_state([state_low[1], temp_low],
                                               "PT", cb.fprop._TRANS_STRING)  # super heated state at low pressure
                s_speed = state_sup[-1]  # needed for the machines
                state_sup = state_sup[:7]
                if (state_low[1] > p_low) and (state_low[1] < p_high):
                    state_low_boil = my_fluid.set_state([state_low[1], 0],
                                                        "PQ")  # saturated liquid at low pressure
                    state_high = my_fluid.set_state(
                        [temp_high, 1.], "TQ")  # find high pressure, sat. vapor

                    if (state_high[1] < p_high and state_high[1] > p_low):

                        state_is = my_fluid.set_state(
                            [state_high[1], state_sup[4]], "PS")  # isentropic compression
                        work_is = state_is[2] - state_sup[2]
                        work_80 = work_is / eff_isentropic
                        # state after compression with eff_isentropic
                        state_is80 = my_fluid.set_state(
                            [state_high[1], state_sup[2] + work_80], "PH")

                        # saturated liquid at high pressure
                        state_dew = my_fluid.set_state(
                            [state_high[1], 0], "PQ")

                        # check the deviation from linearity at high p
                        state_average = (state_high + state_dew) / 2
                        state_mid = my_fluid.set_state(
                            [state_high[1], state_average[2]], "PH")
                        lin_temp_deviation = (state_mid[0] - state_average[0])

                        if temp_limit and state_dew[0] > temp_low:
                            state_throttle = my_fluid.set_state([state_dew[2],
                                                                 state_low[1]], "HP")  # throtteling the liquid to low pressure
                            state_high_p_low__temp = my_fluid.set_state(
                                [state_high[1], temp_low], "PT")  # high pressure cooled to low_t
                            state_throttle_low = my_fluid.set_state([state_high_p_low__temp[2],
                                                                    state_low[1]], "HP")
                            results.append(np.array([*mole_fractions, *state_low,
                                                     *state_sup,
                                                     *state_high,
                                                     *state_is,
                                                     *state_is80,
                                                     *state_dew,
                                                     *state_mid,
                                                     *state_throttle,
                                                     *state_high_p_low__temp,
                                                     *state_throttle_low,
                                                     *state_low_boil]))
                            sound_speeds.append(s_speed)
                            lin_t_deviations_all.append(lin_temp_deviation)
            except Exception as ex_message:
                exception_messages.append(ex_message)

    property_names = cb.fprop._fl_properties_names[:7]
    results = np.array(results)
    names = [*names, *[n+"_l" for n in property_names],
             *[n+"_sup" for n in property_names],
             *[n+"_h" for n in property_names],
             *[n+"_is" for n in property_names],
             *[n+"_is80" for n in property_names],
             *[n+"_dew" for n in property_names],
             *[n+"_mid" for n in property_names],
             *[n+"_thr" for n in property_names],
             *[n+"_hplt" for n in property_names],
             *[n+"_thrlow" for n in property_names],
             *[n+"_bol" for n in property_names]
             ]

    dframe = pd.DataFrame(data=results, columns=names)
    
    good_points =len(results[:,0])
    first_level = ["mole_fractions"]*n_species
    
    point_names = ["low p, x =1","superheated", "high p, x =1", "isentropic",
                   "80%-is", "dewpoint", "midpoint", "throttle", "hplt",
                   "thrlow", "bol"]
    for name in point_names:
        for i_property in range(7):
            first_level.append(name)
    second_level =[*fluids_all,*property_names*len(point_names)]
    dframe_multi = pd.DataFrame(data=results, columns=[first_level, second_level])
    
    # dframe.to_csv(fname+".csv")
    dframe["speed_of_sound_sup"] = sound_speeds
    dframe["p_difference"] = dframe["Pressure_h"] - dframe["Pressure_l"]
    dframe["p_ratio"] = dframe["Pressure_h"] / dframe["Pressure_l"]
    dframe["T_glide_h"] = dframe["Temperature_h"] - dframe["Temperature_dew"]
    dframe["delta_T_mid"] = lin_t_deviations_all
    dframe["dv_v''"] = 1 - (dframe[property_names[3]+"_thr"]
                            + dframe[property_names[3]+"_dew"]) / \
        (dframe[property_names[3]+"_is"] + dframe[property_names[3]+"_sup"])
    if temp_limit:
        # dframe["dv_v''"] = 1 - (dframe[property_names[3]+"_thr"]
        #                         + dframe[property_names[3]+"_dew"]) / \
        #     (dframe[property_names[3]+"_is"] +
        #      dframe[property_names[3]+"_sup"])
        dframe["dv_v''_b"] = 1 - (dframe[property_names[3]+"_thrlow"]  # does it help to subcool?
                                  + dframe[property_names[3]+"_hplt"]) / \
            (dframe[property_names[3]+"_is"] + \
             dframe[property_names[3]+"_sup"])
        dframe["COP_is"] = (dframe[property_names[2]+"_is"]
                            - dframe[property_names[2]+"_dew"]) /\
            ((dframe[property_names[2]+"_is"]-dframe[property_names[2]+"_sup"]))
        dframe["COP_is80"] = (dframe[property_names[2]+"_is80"]
                              - dframe[property_names[2]+"_dew"]) /\
            ((dframe[property_names[2]+"_is80"] -
             dframe[property_names[2]+"_sup"]))

    dframe.to_csv(fname+".csv")
    col_names = dframe.columns
    index_f = 9*["evaluated"]
    index_s = list(col_names[-9:])
    dframe_multi[index_s]=dframe[col_names[-9:]]
    dframe_multi.to_csv(fname+"-multi.csv")
    
    # to remember
    # get only temperatures:
    # df = dframe_multi.loc(axis=1)[:, 'Temperature']
    # idx = pd.IndexSlice
    # df= dframe_multi.loc(axis=1)[idx[:,["Temperature","Pressure"]]]

    # Plot
    figure, axes = plt.subplots(
        figsize=(10, 10), layout="constrained", nrows=1, ncols=1)
    fff = sbn.scatterplot(x="T_glide_h", y="p_ratio",
                          hue=names[0], size=names[1],
                          style=names[2], data=dframe.round(3), ax=axes)
    sbn.move_legend(fff, "upper left", bbox_to_anchor=(1, 1))
    axes.set_title(f"Mixture: {fluids_all}")
    figure.savefig(fname+".png")

    all_results["exception_messages"] = exception_messages
    all_results["results_DataFrame"] = dframe
    all_results["results_DataFrame_multi"] = dframe_multi
    return all_results


def eval_is_eff_roskosch(data, file_out):
    """
    Evalutes the data in the combined dataFrame with the initial fluid screening

    and the output of the Roskosch compressor model (h_aus, s_aus, h_e) for
    exactly the states
    calculated along the screening. The column names are quite special and
    one has to know that the Roskosch model calculates in kJ, while carbatpy
    uses SI units (J). With the output enthalpy of the Roskosch model, the
    **COP_comp** is calculated. Using the enthalpies and entropies along the
    isobaric heat ransfer, mean temperatures are calculated.Here are again two
    cases

    .. line-block::
        a) for throttling at quality=0 
        b) throttling after subcooling to     T_low (names: *_lowT*). 

    With these mean
    temperatures, the COPs for two reversible cases are calculated: for the
    isentropic efficieciecy of 80% *COP_rev80* and for the Roskosch 'real'
    case *COP_rev_r*. Finally, the (pseudo-)real COP is compared to the
    reversible COP, giving a second law efficiency *eff_sec_law_r* for the
    Roskosch efficiencies and *eff_sec_law_80* for the fixed 80% efficiency,
    which include the
    compressor and the throttling, but **no heat transfer**!

    The Roskosch piston compressor model is described here:
    http://dx.doi.org/10.1016/j.ijrefrig.2017.08.011

    Is part of carbatpy.

    Parameters
    ----------
    data : pandas.dataFrame
        as calculated by combining the fluid screening dataFrame with the
        ouput of the Roskosch model (as dataFrame).
    file_out : string
        name of the file (incl. directory) where the resulting dataFrame shall
        be stored.

    Returns
    -------
    data : pandas.dataFrame
        input expanded by the results.

    """

    # mean low T, first with throttling at x=0, then with subcooling to T_low
    # index name for the latter _lowT
    data["T_mean_low"] = (data['spec_Enthalpy_sup'] - data['spec_Enthalpy_thr']
                          ) / (data['spec_Entropy_sup'] - data['spec_Entropy_thr'])
    data["T_mean_low_lowT"] = (data['spec_Enthalpy_sup'] - data['spec_Enthalpy_thrlow']
                               ) / (data['spec_Entropy_sup'] - data['spec_Entropy_thrlow'])

    try:  # if the Roskosch-model data are in the dataFrame
        data["COP_comp"] = (data["h_aus"] * 1000 - data['spec_Enthalpy_dew']) /\
            ((data["h_aus"] - data["h_e"]) * 1000)
        data["T_mean_high_is_r"] = (
            data['h_aus'] * 1000 - data['spec_Enthalpy_dew'])/(data['s_aus'] - data['spec_Entropy_dew'])
        data["COP_rev_r"] = data["T_mean_high_is_r"] / (data["T_mean_high_is_r"]
                                                        - data["T_mean_low"])
        data["COP_rev_r_lowT"] = data["T_mean_high_is_r"] / (data["T_mean_high_is_r"]
                                                             - data["T_mean_low_lowT"])
        data["eff_sec_law_r"] = data["COP_comp"] / data["COP_rev_r"]
        data["eff_sec_law_r_lowT"] = data["COP_comp"] / data["COP_rev_r_lowT"]
    except:
        pass

    # values for isentropic efficiency of 80% and throttling at a quality of 0
    data["T_mean_high_is80"] = (data['spec_Enthalpy_is80'] - data['spec_Enthalpy_dew']) \
        / (data['spec_Entropy_is80'] - data['spec_Entropy_dew'])
    data["COP_rev80"] = data["T_mean_high_is80"] / (data["T_mean_high_is80"]
                                                    - data["T_mean_low"])
    data["eff_sec_law_80"] = data["COP_is80"] / data["COP_rev80"]

    # values for isentropic efficiency of 80% and throttling at the low T and high p
    data["T_mean_high_is80_lowT"] = (data['spec_Enthalpy_is80'] - data['spec_Enthalpy_hplt']) \
        / (data['spec_Entropy_is80'] - data['spec_Entropy_hplt'])
    data["COP_rev80_lowT"] = data["T_mean_high_is80"] / (data["T_mean_high_is80"]
                                                         - data["T_mean_low_lowT"])
    data["eff_sec_law_80_lowT"] = data["COP_is80"] / data["COP_rev80_lowT"]
    data.to_csv(file_out)
    return data


def get_fluid(data):
    """
    find the fluids used in the screening

    Parameters
    ----------
    data : pandas.dataFrame
        dataFrame from fluid screening with mole fractions. In the column names
        the names of the fluids are found.

    Returns
    -------
    fluids : list of strings
        names of the fluids.
    fluid_col : list of strings
        List with the column names (includes a sarting"x_".
    fluid_str : string
        Fluid composition string as accepted by RefProp.

    """
    fluids = []
    fluid_col = []
    col_names = data.columns
    for name in col_names:
        if name.find("x_") > -1:
            if name not in fluid_col:  # no doubles
                fluids.append(name[2:])
                fluid_col.append(name)
    fluid_str = "*".join(fluids)
    return fluids, fluid_col, fluid_str


def combine(filenames, filename_out="automatic"):
    """
    Combine two data frames out of two or more files with same number of lines

    and fitting to each other. Can be used, when after fluid screening
    machine efficienceies, costs,
    etc. are calculated as post-processing. Can help in evaluation and
    plotting.

    Parameters
    ----------
    filenames : list of strings
        all filenames (incl. directories), to be read.
    filename_out : string, optional
        Where to store the result. The default is "automatic". Then the first
        filename isxpanded vy "-combined".

    Raises
    ------
    ValueError
        If tgere is a problem with the files.

    Returns
    -------
    combined : pandas.dataFrame
        the combined dataFrame.

    """
    all_frames = []
    if filename_out == "automatic":
        fname_new = filenames[0].split(".")
        filename_out = fname_new[0]+"-combined0." + fname_new[1]
    try:
        for which in filenames:
            all_frames.append(pd.read_csv(which))

    except Exception as excep:
        text = f"{which} vs. {os.getcwd()}, {excep}"
        raise ValueError(text) from excep
    combined = pd.concat(all_frames, axis=1)
    combined.to_csv(filename_out)
    return combined



def data_plot(filename, what, filename_out="automatic", fig_title=""):
    """
    Plotting a dataframe from a file using a dictionary with the keys

    being the plotted parameters "x", "y", "hue" etc. and storing it to
    a file.

    Parameters
    ----------
    filename : string
        csv-file with the data-frame to be imported.
    what : dictionary
        keys "y","x","hue", "style", size etc. values must be some column names. .
    filename_out : string, optional
        where to store the plot, including directory. The default is "automatic".
    fig_title : string, optional
        Title of the figure to be plotted, default is "".

    Returns
    -------
    bool
        success?

    """
    try:
        dframe = pd.read_csv(filename)

        set(what).issubset(set(dframe.columns))
    except Exception as excep:
        print(f"{what} not in columns!\n{excep}")
        return False
    # Plot
    figure, axes = plt.subplots(
        figsize=(10, 10), layout="constrained", nrows=1, ncols=1)
    fff = sbn.scatterplot(x=what["x"], y=what["y"],
                          hue=what["hue"], size=what["size"],
                          style=what["style"], data=dframe.round(3), ax=axes)
    sbn.move_legend(fff, "upper left", bbox_to_anchor=(1, 1))
    axes.set_title(fig_title)
    if filename_out == "automatic":
        filename_out = filename.split(".")[0] + "-plot2.png"
    figure.savefig(filename_out)
    return True

def plot_cycle(filename, dataset):
    try:
        dframe = pd.read_csv(filename)

        
    except Exception as excep:
        print(f"Problem: {excep}")
        return False
    
    
def get_cycle_points(data, index):
    
    indices =range(7)
    

    sup_names = [ "_l" ,
             "_sup" ,
             "_h" ,
             "_is" ,
             "_is80" ,
             "_dew" ,
             "_mid" ,
             "_thr" ,
             "_hplt" ,
             "_thrlow" ,
             "_bol" 
             ]
    n_points = len(sup_names)
    points =np.zeros((len(indices),n_points))
    for outer, variable in enumerate(indices):
        property_name = cb.fprop._fl_properties_names[variable]
        names = [property_name + sup for sup in sup_names]
        
        
        points[outer,:] = data.loc[index, names]
    return points

if __name__ == "__main__":
    FLUIDS_ACTUAL = ["Propane", "Ethane", "Pentane",
                     "Butane"]  # ["DME", "Ethane", "Butane","CO2"]
    TEMP_LOW = 285.00
    TEMP_HIGH = 363.00
    PRESSURE_LOW = 10E4
    PRESSURE_HIGH = 22E5
    DIRECTORY_NAME = cb._RESULTS_DIR + r"\optimal_hp_fluid\fluid_select_restricted"
    TEMPERATURE_LIMIT = True
    warn = mixture_search(FLUIDS_ACTUAL, [TEMP_LOW, TEMP_HIGH],
                          [PRESSURE_LOW, PRESSURE_HIGH],
                          DIRECTORY_NAME, resolution=21,
                          temp_limit=TEMPERATURE_LIMIT)
    #####################################################

    directory = cb._CARBATPY_BASE_DIR
    directory += "\\tests\\test_files\\"

    filename1 = directory + r"test_data_ProEthPenBut\2024-02-06-16-51-ProEthPenBut.csv"
    filename2 = directory + \
        r"test_data_ProEthPenBut\2024-02-06-16-51-ProEthPenBut-compressor-Roskosch.csv"
    combined_data = cb.utils.property_eval_mixture.combine([filename1,
                                                            filename2],
                                                           filename_out="automatic")
    ##############################################
    what_act = {"x": 'spec_Volume_sup', "y": 'COP_is80', "hue": 'T_glide_h',
                "size": 'p_ratio', 'style': 'Temperature_hplt'}
    SUCCESS = data_plot(filename1, what_act)

    ############################################
    fluids_act, fluid_col_act, fluid_str_act = get_fluid(combined_data)

    #########################################
    evaluated_data = eval_is_eff_roskosch(combined_data,
                                          directory+'evaluated.csv')



# -*- coding: utf-8 -*-
"""
How does the uncertainty in spec. volume influence the uncertainty in enthalpy?

A Monte Carlo approach

Created on Wed Nov 15 10:37:45 2023

@author: atakan
"""
import numpy as np
import numpy.polynomial.polynomial as pol

#import numpy.random.Generator as nrg
import matplotlib.pyplot as plt

import carbatpy as cb

_RESULTS_ = cb._RESULTS_DIR


# FLUID = "Propane * Ethane * Pentane *Butane"
# #comp = [.75, 0.05, 0.15, 0.05]
# x1 = 0.
# x2 = 0.3
# x3 = 0.2
# x4 = 1 - x1-x2-x3
# comp = [x1, x2, x3, x4]  # [0.164,.3330,.50300,0.0]

###############
FLUID = "Butane * CO2"
comp = [0.8, 0.2]
print(f"{FLUID}, composition:\n{comp}")


flm = cb.fprop.FluidModel(FLUID)
myFluid = cb.fprop.Fluid(flm, comp)
temp_low = 300.
delta_temp = 10
n_points = 21
pressure_ratios = np.linspace(1, 11, n_points)
state_in = myFluid.set_state([temp_low, 1.], "TQ")
p_low = state_in[1]
state_in = myFluid.set_state([temp_low + delta_temp, p_low], "TP")
results = []
entropy = state_in[4]
for p_out in pressure_ratios:
    results.append(myFluid.set_state([p_out*p_low, entropy], "PS"))

results = np.array(results)
h_initial = np.trapz(results[:,3], results[:,1])
f, ax = plt.subplots(2, 2)
# fit a 'polytropic' [degree: 2] to the Refprop data log(v) vs log(p)
para = pol.polyfit(np.log10(results[:, 1]), np.log10(results[:, 3]), 2)
const = results[0, 1]* results[0, 3]**(-para[1])
ax[0,0].plot(results[:, 1], results[:, 3], "x")
#ax[0].plot(results[:, 1], (const  /results[:, 1])**(-1/para[1]))


# ax[0].plot(np.log10(results[:, 1]), np.log10(results[:, 3]), "x")
ax[0,0].plot(results[:, 1], 10**pol.polyval(np.log10(results[:, 1]), para))

para_neu =para

ax[0,0].plot(results[:, 1], 10**pol.polyval(np.log10(results[:, 1]), para))

# Now some scatter, two methods a) scatter on each v-point [middle plot] and
# b) scatter on 3 points (see indx) and fit a polytropic through these points [right plot]
# and evaluate the intergral for each appoach
np.random.normal()
dh = [] # approach a)-Results
dh_alternative = []  # approach b) results
dh_shift =[]  # the whole curve is shifted multiplied by y single :.3gfactor
indx = [0, n_points//2, -1]
order = 2
if len(indx) < 3:
    order = 1
for sample in range(1000):
    factors = np.random.normal(1, 0.08, n_points)
    f2 = factors[indx]
    pres_2 = pressure_ratios[indx]*p_low
    para2 = pol.polyfit(np.log10(pres_2), np.log10(results[indx, 3]*f2), order)
    p = pressure_ratios*p_low
    # volumes = 10**pol.polyval(np.log10(p), para)*factors
    volumes = results[:,3]*factors
    volumes2 = 10**pol.polyval(np.log10(p), para2)# *factors
    # volumes3 = factors[0]* 10**pol.polyval(np.log10(p), para)
    volumes3 = results[:,3]*factors[0]
    ax[0,1].plot(p, volumes)
    ax[1,1].plot(p, volumes2)
    ax[1,0].plot(p, volumes3)
    dh.append(np.trapz(volumes, p))
    dh_alternative.append(np.trapz(volumes2, p))
    dh_shift.append(np.trapz(volumes3, p))
dh = np.array(dh)
dh_alternative = np.array(dh_alternative)
dh_shift = np.array(dh_shift)

f.savefig(_RESULTS_+"v_uncertain.png")
print (f"Initial delta h: {h_initial:.2f}")
print(f"independent scatter on every point, mean {dh.mean():.2f}, std-dev {dh.std():.2f}, relative {dh.std()/dh.mean():.2f}")
print(f"scatter on 3 points &  fit, mean {dh_alternative.mean():.2f}, std-dev { dh_alternative.std():.2f}, relative {dh_alternative.std()/dh_alternative.mean():.2f}")
print(f" same scatter on all points &  fit, mean {dh_shift.mean():.2f}, std-dev {  dh_shift.std():.2f}, relative {dh_shift.std()/dh_shift.mean():.2f}")



# -*- coding: utf-8 -*-
"""
Created on Mon May 22 21:09:12 2023

@author: atakan
"""

import os
# from time import time
from ctREFPROP.ctREFPROP import REFPROPFunctionLibrary
import numpy as np
# import CoolProp.CoolProp as CP
import carbatpy
if carbatpy._TREND["USE_TREND"]:
    try:
        import fluid as tr_fl  # TREND fluids
    except ImportError as e:
        print(f"Import error for 'fluid': {e}")
    


# os.environ['RPPREFIX'] = r'C:/Program Files (x86)/REFPROP'
# os.environ['RPPREFIXs'] = r'C:/Program Files (x86)/REFPROP/secondCopyREFPROP'
_PROPS = "REFPROP"  # or "CoolProp"

_fl_properties_names = ("Temperature", "Pressure", "spec_Enthalpy",
                        "spec_Volume", "spec_Entropy", "quality",
                        "spec_internal_Energy",
                        "viscosity", "thermal_conductivity",
                        "Prandtl_number", "k_viscosity", "molecular_mass",
                        "speed_of_sound")
_THERMO_STRING = "T;P;H;V;S;QMASS;E"
_THERMO_TREND = request = ["T","P","H" "D", "S", "Q","U"] # careful density not volume
_TRANS_STRING = _THERMO_STRING + ";VIS;TCX;PRANDTL;KV;M;W"
_TRANS_TREND = _THERMO_TREND.extend([ "ETA","TCX", "WS"])
_TV_STRING = "T;V"
_T_SURROUNDING = 288.15 # K
_MODEL_ARGS = {}

# order for coolprop,alle_0:[_temp, p,  h, 1/ rho, s,x,cp, mu,  lambda_s,
# prandtl, phase]"
_UNITS = 21
rp_instance = ""
if _PROPS == "REFPROP":
    try:
        rp_instance = REFPROPFunctionLibrary(os.environ['RPPREFIX'])
        # be careful pressure is in Pa!
        _UNITS = rp_instance.GETENUMdll(0, "MASS BASE SI").iEnum
    except:
        print("print refprop not installed!")
elif _PROPS == "TREND":
    pass
else:
    print(f"_PROPS value not unstalled {_PROPS}")



class FluidModel:

    def __init__(self, fluid, units=_UNITS, props=_PROPS, rp_inst=rp_instance,
                 args=_MODEL_ARGS):
        """
        For Class Fluidmodell a fluid (mixture) must be defined, the evaluation
        takes place with props, units can be set and an instance can be set,
        the latter is important, if more than one fluid is used.

        Parameters
        ----------
        fluid : String
            as defined in the props (Model), for RefProp it is
            "fluid1 * fluid2 * fluid3".
        units : integer, optional
            units, normally SI wuth MASS base, see props(RefProp).
            The default is _UNITS.
        props : string, optional
            select the property model. The default is _PROPS.
        rp_inst : RefProp-Instance, optional
            where Refprop is installed, for two fluids, as in heat exchangers,
            two installations/instances are needed. The default is rp_instance.

        Returns
        -------
        None.

        """
        self.fluid = fluid
        self.props = props
        self.units = units
        self.args = args
        if props == "REFPROP":
            self.rp_instance = rp_inst
            self.set_rp_fluid()
        elif props == "TREND":
            self.set_tr_fluid()

    def set_rp_fluid(self, modwf=REFPROPFunctionLibrary, name='RPPREFIX'):
        """
        A new instance of Refpropdll for the given fluid. It can then be called
        using fluid =""

        Parameters
        ----------
        fluid : string
            fluid (mixture) name, as described in REFPROP.

        Returns
        -------
        self.rp_instance : REFPROP Instance
            for further usage.

        """

        self.rp_instance = modwf(os.environ[name])
        self.rp_instance.SETPATHdll(os.environ[name])
        ierr = self.rp_instance.SETFLUIDSdll(self.fluid)
        if ierr != 0:
            print(f"Fehler in setfluid {ierr}")
            print(self.rp_instance.ERRMSGdll(ierr))
        return self.rp_instance
    
    def set_tr_fluid(self):
        self.fluid = self.fluid_to_list()
        """
        in_states, 
        fld = fluid.Fluid(self.args["input"],
                          self.args['output'],
                          species,
                          self.["composition"],
                          self.
            'TP','H',['water'],[1],[1],1,path,'specific',dll_path)

        
        """
        
        
    def fluid_to_list(self):
        no_blank = self.fluid.replace(" ","")
        return no_blank.split("*")
        
        


class FluidState:
    def __init__(self, state):
        self.temperature = state.Output[0]
        self.pressure = state.Output[1]
        self.sp_volume = state.Output[3]
        self.enthalpy = state.Output[2]
        self.entropy = state.Output[4]
        self.quality = state.Output[5]
        self.int_energy = state.Output[6]
        self.state = state.Output[:7]
        self.prop_names = _fl_properties_names


class FluidStateTV:
    def __init__(self, state):
        self.temperature = state.Output[0]
        self.sp_volume = state.Output[1]
        self.state = state.Output[:1]


class FluidStateTransport(FluidState):
    def __init__(self, state):
        super().__init__(state)
        # ";VIS;TCX;PRANDTL;KV;M;W"
        self.viscosity = state.Output[7]
        self.thermal_conductivity = state.Output[8]
        self.prandtl = state.Output[9]
        self.kin_viscosity = state.Output[10]
        self.molecular_mass = state.Output[11]
        self.speed_of_sound =state.Output[12]
        self.transport = state.Output[7:13]
        self.state = state.Output[:13]


class Fluid:
    """ 
    The Fluid class is used to set, get, and print states of  a fluid with a given
    
    model (e.g. RefProp). The compounds are set in the fluidmodel, while the
    composition is also set here.
    
    """

    def __init__(self, fluidmodel, composition=[1.0],
                 option=1):
        self.fluidmodel = fluidmodel
        self.composition = composition
        self.option = option
        self.no_compounds = len(composition)
        self.herr = 0

    def set_composition(self, composition):
        self.composition = composition
        if self.fluidmodel.props == "TREND":
            # self.comp
            pass # here the mass fraction to internal mol fraction conversion
        
    

    def set_state(self, values, given="TP",
                  wanted=_THERMO_STRING,
                  composition=[]):
        if composition == []:
            composition = self.composition
        else:
            self.composition = composition

        if self.fluidmodel.props == "REFPROP":
            state = self.fluidmodel.rp_instance.REFPROP2dll(
                self.fluidmodel.fluid, given, wanted,
                self.fluidmodel.units,
                0, values[0], values[1],
                composition)

            if state.ierr == 0:
                if wanted == _THERMO_STRING:
                    # print(state)
                    self.properties = FluidState(state)
                elif wanted == _TRANS_STRING:
                    self.properties = FluidStateTransport(state)
                elif wanted == _TV_STRING:
                    self.properties = FluidStateTV(state)
                else:

                    raise Exception(f"properties{wanted} not implemented yet!")
            else:
                self.herr = state.herr
                raise Exception(f"Property-Refprop problem: {state.herr}!")
        else:
            raise Exception(
                f"Property model {self.fluidmodel.props} not implemented yet!")
        return np.array([*self.properties.state])

    def set_state_v(self, values, given="TP", wanted=_THERMO_STRING):
        dimension = np.shape(values)
        number_wanted = wanted.count(";")+1
        output = np.zeros((dimension[0], number_wanted))
        for count, value in enumerate(values):
            output[count, :] = self.set_state(value, given, wanted)
        self.state_v = output
        return output

    def print_state(self):
        pr = self.properties
        flm = self.fluidmodel
        print(f"\n{flm.fluid}, composition: {self.composition}")
        print(f"T:{pr.temperature:.2f} K, p: {pr.pressure/1e5 :.2f} bar,  h: {pr.enthalpy/1000: .2f} kJ/kg, s: {pr.entropy/1000:.3f} kJ/kg K\n")
        if pr.quality >= 0:
            print(f"Quality: {pr.quality :.3f}")
            
    def calc_temp_mean(self, h_final):
        """
        Calculate the thermodynamic mean temperature between the actual state
        
        and the final enthalpy along an isobaric (Delta h /Delta s)

        Parameters
        ----------
        h_final : float
            enthalpy of he final stat.

        Returns
        -------
        temp_mean : float
            the thermodynamic mean temperature.

        """
        actual_props = self.properties.state
        final_props = self.set_state([h_final, actual_props[1]], "HP")
        temp_mean = (final_props[2] - actual_props[2]) \
            / (final_props[4] - actual_props[4]) 
        return temp_mean
    
    
def init_fluid(fluid, composition, **keywords):
    """
    short way to define a Fluid and a FluidModel

    Parameters
    ----------
    fluid : string
        The species within the fluid.
    composition : List
        mole fraction for each fluid.
    **keywords : TYPE
        all keywords needed for the FluidModel, if non-defults shall be set.

    Returns
    -------
    actual_fluid : Fluid
        Instance of the actually set Fluid.

    """
    flm = FluidModel(fluid, **keywords)
    actual_fluid = Fluid(flm, composition)
    return actual_fluid

if __name__ == "__main__":
    FLUID = "Propane * Pentane"
    comp = [.50, 0.5]
    flm = FluidModel(FLUID)
    myFluid = Fluid(flm, comp)
    st0 = myFluid.set_state([300., 1e5], "TP")
    st1 = myFluid.set_state([300., 1e5], "TP",
                            _TRANS_STRING)
    print(st0, st1)
    myFluid.print_state()
    myFluid.set_composition([.2, .8])
    st0 = myFluid.set_state([300., 1e5], "TP", composition=[.35, .65])
    myFluid.print_state()
    
    mean_temp_act = myFluid.calc_temp_mean(st0[2]+1e5)
    print(f"Mean Temperature {mean_temp_act} K")

    # value_vec = np.array([[300, 1e5], [400, 1e5], [500, 1e5]])
    # stv = myFluid.set_state_v(value_vec, "TP")

    # print(myFluid.set_state_v(value_vec, "TP"))
    # print(myFluid.set_state([300., 1.], "TQ"))
    
    # New simple way to get an instance of Fluid
    myFluid2 = init_fluid( FLUID, comp)



# -*- coding: utf-8 -*-
"""
Created on Sun May 21 10:52:09 2023

@author: atakan
"""

import copy
import numpy as np
import matplotlib.pyplot as plt


import carbatpy as cb
from carbatpy import _RESULTS_DIR
_RESULTS_ = _RESULTS_DIR


class OrganicRankineCycle:
    """
    Heat pump class for simple thermodynamic calculations for energy storage

    Two pairwase storages at two different temperatures are assumed,
    always starting at room temperature (around 300K). Thus, three fluids are
    expected: the working fluid, the storage fluid  or sink at high temperature
    (h) and the source storage fluid at low temperature.
    The minimum and maximum temperatures of all
    storags are fixed and the heat flow rate to the high-T storage are
    given. Also, isentropic efficiencies are fixed so far, but they should use
    a function later.
    """

    def __init__(self,  fluids, fixed_points, components=[]):
        """
        Heat pump class for simple thermodynamic calculations for energy storage

        storages at two different temperatures, always starting at room
        temperature (around 300K). The minimum and maximum temperatures aof all
        storags are fixed and the heat flow rate to the high-T storage are
        given. Also, isentropic efficiencies. The COP_charging is important
        to reach a steady state. At the moment you will have to look it up from
        the heat pump calculation. It is used to divide the energy transfered to
        the envoronment and to the low temperature storage.
        Example for::

        fixed_points = {"eta_s": _ETA_S_,  # expander
                        "eta_s_p": _ETA_S_P_,  # pump
                        "p_low": p_low,
                        "p_high": p_high,
                        "T_hh": _STORAGE_T_IN_,
                        "h_h_out_sec": state_sec_out[2],
                        "h_h_out_w": state_out_evap[2],
                        "h_l_out_cold": state_cold_out[2],
                        "h_l_out_w": state_out_cond[2],
                        "h_env_in": state_env_in[2],
                        "h_env_out": state_env_out[2],
                        "T_hl": _STORAGE_T_OUT_,
                        "T_lh": _COLD_STORAGE_T_OUT_,
                        "T_ll": _COLD_STORAGE_T_IN_,  # 256.0,
                        "Q_dot_h": _Q_DOT_MIN_,
                        "d_temp_min": _D_T_MIN_,
                        "cop_charging": _COP_CHARGING  # needed to calculate Q_env_discharging
                        }

        Parameters
        ----------
        fluids : list of three Fluid
            working fluid, secondary fluid to store at high T, cold fluid
            to store at low T.
        fixed_points : Dictionary
            all the fixed points are summarized here, the isentropic efficiency,
            the lower pressure(evaporator), the high and low temperatures of
            the high temperature storages, the same for the low temperature
            storages(source), the heat to be transfered at high T, the minimum
            approach temperature.
        components : TYPE, optional
            DESCRIPTION. The default is [].

        Returns
        -------
        None.

        """
        self.components = components
        self.fixed_points = fixed_points
        self.fluids = fluids
        self.all_states = []
        self.m_dots = []
        self.evaluation = {"Q_dot_h": self.fixed_points["Q_dot_h"],
                           "Power-Net": 0.0,
                           "T_hh": self.fixed_points["T_hh"],
                           "T_ll": self.fixed_points["T_ll"],
                           #"exergy_loss_rate": 0,
                           "eta_pc": 0}
        self.warning = []

    def calc_orc(self,  verbose=False):
        """
        calculate a simple organic rankine cycle (ORC)

        working with two sensible storages (or source and sink). At the moment
        the machine model is just for a constant isentropic efficiency.
        The initial states are provided
        by the Fluid-instances. The pressures have to be selected carefully,
        they are not tested at the moment.
        For optimization, the pressure levels of the cycle could be used within
        a second script. Several parameters are fixed
        along a calculation, this is in the dictionary "self.fixed_points".


        Parameters
        ----------
        verbose : Boolean, optional
            should values be printed? The deault is False

        Returns
        -------
        eta_orc : flaot
            thermal efficiency of ORC.


        """

        state_in = copy.copy(self.fluids[0])  # ORC starts before the pump
        w_actual = self.fluids[0]
        h_initial = w_actual.properties.enthalpy
        if verbose:
            print("Start:")
            w_actual.print_state()

        # pump------------------------
        w_after_pump, work_pump, _power = \
            cb.compressor_simple.compressor(self.fixed_points["p_high"],
                                            self.fixed_points["eta_s_p"],
                                            self.fluids[0]
                                            )

        # the total energy available is given, the mass flow rate of the
        # working fluid should be calculated first.
        m_dot_wf = self.fixed_points["Q_dot_h"] / (self.fixed_points["h_h_out_w"]
                                                   - w_actual.properties.enthalpy)
        # absolute value!
        power_pump = work_pump * m_dot_wf
        if verbose:
            print(f"pump-power {power_pump:.2f} W, m_dot: {m_dot_wf:.4f} kg/s")
            w_actual.print_state()
        evaporator = cb.hex_th.StaticHeatExchanger([w_actual, self.fluids[1]],
                                                   self.fixed_points["Q_dot_h"],
                                                   self.fixed_points["h_h_out_w"],
                                                   self.fixed_points["h_h_out_sec"],
                                                   d_temp_separation_min=self.fixed_points["d_temp_min"],
                                                   calc_type="const",
                                                   name="evaporatorORC")
        result_ev = evaporator.find_pinch(False)
        ms_ev, d_tempall_ec, w_ev, s_ev = evaporator.pinch_plot(
            plotting=verbose)
        if evaporator.warning < 100:
            self.evaluation[evaporator.name] = {
                "dT-min": evaporator.dt_min,
                "dT-mean":  evaporator.dt_mean,
                "Q-dot":  evaporator.q_dot,
                "Warning": evaporator.warning}
        else:
            self.warning.append([evaporator.name, evaporator.warning])

        self.all_states.append(w_ev)
        self.all_states.append(s_ev)
        self.m_dots.append(m_dot_wf)
        self.m_dots.append(ms_ev)
        w_after_evaporator = self.fluids[0].set_state(
            [w_ev[0][2], w_ev[0][1]], "HP")

        if verbose:
            print("after evaporator:")
            w_actual.print_state()

        # expander -----------------------
        w_after_expander, work_expander, _power =\
            cb.compressor_simple.compressor(self.fixed_points["p_low"],
                                            self.fixed_points["eta_s"],
                                            self.fluids[0]
                                            )
        power = work_expander * m_dot_wf  # absolute value!

        if verbose:
            print(f"After expander, power: {power : .2f}")
            w_actual.print_state()

        # condenser 1: environment------------------------
        q_dot_l = self.fixed_points["Q_dot_h"] * \
            (1-1/self.fixed_points["cop_charging"])
        q_dot_env = self.fixed_points["Q_dot_h"] + power_pump + power - q_dot_l
        h_wf_out_env = w_after_expander[2] - q_dot_env/m_dot_wf
        condenser_env = \
            cb.hex_th.StaticHeatExchanger([w_actual,
                                           self.fluids[2]],
                                          q_dot_env,
                                          h_wf_out_env,
                                          self.fixed_points["h_env_out"],
                                          d_temp_separation_min=self.fixed_points["d_temp_min"],
                                          name="condenserORC_env")
        h_env_out = condenser_env.find_pinch()
        m_dot_envf, d_tempall, wf_states, envf_states =\
            condenser_env.pinch_plot(plotting=verbose)
        if condenser_env.warning < 100:
            self.evaluation[condenser_env.name] = {
                "dT-min": condenser_env.dt_min,
                "dT-mean": condenser_env.dt_mean,
                "Q-dot": condenser_env.q_dot,
                "Warning": condenser_env.warning}
        else:
            self.warning.append([condenser_env.name, condenser_env.warning])

        self.all_states.append(wf_states)
        self.all_states.append(envf_states)
        self.m_dots.append(m_dot_wf)
        self.m_dots.append(m_dot_envf)
        w_after_cond_1 = self.fluids[0].set_state(
            [wf_states[0][2], wf_states[0][1]], "HP")
        if verbose:
            print("After first condenser")
            w_actual.print_state()

        # Now the second condenser is needed  BA 2023-11-28
        condenser = \
            cb.hex_th.StaticHeatExchanger([w_actual,
                                           self.fluids[3]],
                                          q_dot_l,
                                          h_initial,
                                          self.fixed_points["h_l_out_cold"],
                                          d_temp_separation_min=self.fixed_points["d_temp_min"],
                                          name="condenserORC")
        # h_cond_out = condenser.find_pinch() # no degree of freedom left!
        m_dot_cold, d_tempall, wf_cond_states, coldf_states =\
            condenser.pinch_plot(plotting=verbose)
        if condenser.warning < 100:
            self.evaluation[condenser.name] = {
                "dT-min": condenser.dt_min,
                "dT-mean":   condenser.dt_mean,
                "Q-dot":  condenser.q_dot,
                "Warning": condenser.warning}
        else:
            self.warning.append([condenser.name, condenser.warning])
        self.all_states.append(wf_cond_states)
        self.all_states.append(coldf_states)
        self.m_dots.append(m_dot_wf)
        self.m_dots.append(m_dot_cold)

        w_after_condenser = self.fluids[0].set_state([wf_cond_states[0, 1],
                                                      wf_cond_states[0, 2]], "PH")
        if verbose:
            print("after condenser", w_after_condenser)
            w_actual.print_state()
            state_in.print_state()
            q_l_alt = m_dot_wf * (w_after_cond_1-w_after_condenser)[2]
            print(f"q_l comparison {q_dot_l:.2f}, {q_l_alt:.2f}")

        eta_orc = np.abs(power + power_pump)/self.fixed_points["Q_dot_h"]
        self.evaluation["Power-Net"] = np.abs(power + power_pump)
        self.evaluation["Power-Pump"] = np.abs(power_pump)
        self.evaluation["eta_pc"] = eta_orc
        return eta_orc

    def hp_plot(self, f_name=_RESULTS_+"\\last_T_H_dot_plot_orc", fig_ax=[]):
        """
        plots the heat pump cycle and stores it to the given file (name)


        Parameters
        ----------
        f_name : string, optional
            where to write the results and the plot.
        fig_ax : list (length 2), optional
            instances with figure and axes to plot into. If Empty a new one will
            be generated. default =[]
        Returns
        -------
        None.

        """
        states = []
        connect_temp = []
        connect_h = []
        position = [0, -1]
        if len(fig_ax) == 0:
            fig, ax = plt.subplots(1, 1)
        else:
            fig, ax = fig_ax
        points = ["-k", "or", "-b", "vg", "-r", "xg"]
        m_dots = self.m_dots
        n_val = [2, 0]  # positions of h and T
        n_points = np.shape(self.all_states)[1]
        n_range = np.shape(m_dots)[0]  # how many streams? 2 x heat exchangers
        data = np.zeros((n_points, 4))
        names = ["ORC-process",  "{$H_{dot} /W", "T / K", "$m_{dot}"]
        all_data = []

        # m_dots[3] = m_dots[3]  # / m_dots[2] * m_dots[0]
        # m_dots[4] = m_dots[2] = m_dots[0]
        h_shift = 0

        # count backwards to shift the env-hex
        for i_process in range(n_range-1, -1, -1):
            h_shift_here = 0
            if i_process in [2, 3]:  # BA This is specific for this cycle
                h_shift_here = h_shift

            cycle_states = self.all_states[i_process]
            cycle_states = cycle_states[cycle_states[:, 2].argsort(), :]
            if i_process == 4:  # BA This is specific for this cycle
                liquid = self.fluids[0].set_state([cycle_states[0, 1], 0.],
                                                  "PQ")
                ax.plot((liquid[2] - cycle_states[0, 2]) * m_dots[i_process],
                        liquid[0], "o")
            cycle_states[:, 2] -= cycle_states[0, 2]

            cycle_states[:, 2] = cycle_states[:, 2] * \
                m_dots[i_process] + h_shift_here
            states.append(cycle_states)

            if i_process == n_range-1:
                h_shift = cycle_states[-1, 2]

            ax.plot(cycle_states[:, 2], cycle_states[:, 0], points[i_process])
            data[:, 0] = float(i_process)
            data[:, -1] = m_dots[i_process]
            data[:, 1:3] = cycle_states[:, n_val]
            all_data.append(copy.copy(data))

            # pump

            # if i_process in ( 0,2,):  # BA This is specific for this cycle
            #     connect_temp.append(cycle_states[position, 0])
            #     connect_h.append(cycle_states[position, 2])
            #     ax.plot(connect_h, connect_temp, "k:")
            # if i_process in ( 0,4,):  # BA This is specific for this cycle
            #     connect_temp.append(cycle_states[position, 0])
            #     connect_h.append(cycle_states[position, 2])
            #     ax.plot(connect_h, connect_temp, "k:")

        # temperatures = np.array(connect_temp).reshape((4))
        # enthalpies = np.array(connect_h).reshape((4))
        ax.set_xlabel("$\dot H$ / W")
        ax.set_ylabel("$T$ / K")
        ax.plot([all_data[5][0][1], all_data[1][0][1]],
                [all_data[5][0][2], all_data[1][0][2]], "b:")
        ax.plot([all_data[5][-1][1], all_data[3][-1][1]],
                [all_data[5][-1][2], all_data[3][-1][2]], "k:")

        # for i_process in range(2):
        #     ax.plot([enthalpies[0+i_process], enthalpies[2+i_process]],
        #             [temperatures[0+i_process], temperatures[2+i_process]], "r:")
        fig.savefig(f_name+".png")
        np.savetxt(f_name+".csv", np.concatenate(all_data),
                   delimiter=";", header=";".join(names))


if __name__ == "__main__":

    FLUID = "Propane * Butane * Pentane * Hexane"
    comp = [.75, 0.05, 0.15, 0.05]
    # comp = [0.4,	0.3,	0.3, 0.0]  # [0.164,.3330,.50300,0.0]

    FLS = "Water"  # Storage fluid
    FLCOLD = "Methanol"  # Storage fluid for low T
    FLENV = "Water"  #

    flm = cb.fprop.FluidModel(FLUID)
    myFluid = cb.fprop.Fluid(flm, comp)

    secFlm = cb.fprop.FluidModel(FLS)
    secFluid = cb.fprop.Fluid(secFlm, [1.])

    coldFlm = cb.fprop.FluidModel(FLCOLD)
    coldFluid = cb.fprop.Fluid(coldFlm, [1.])

    envFlm = cb.fprop.FluidModel(FLENV)
    envFluid = cb.fprop.Fluid(secFlm, [1.])

    # Condenser(c) and storage (s), secondary fluids fix all, temperatures(T in K),
    # pressures (p in Pa)
    _ETA_S_ = 0.67  # interesting when changed from 0.69 to 0.65, the efficiency
    # decreases, the reason is the low quality along throtteling then
    _ETA_S_P_ = 0.6  # pump
    _STORAGE_T_OUT_ = cb._T_SURROUNDING
    _COLD_STORAGE_T_OUT_ = cb._T_SURROUNDING
    _ENV_T_IN_ = cb._T_SURROUNDING
    _ENV_T_OUT_ = cb._T_SURROUNDING + 5.
    _STORAGE_T_IN_ = 363.  # 395.0
    _COLD_STORAGE_T_IN_ = 260.15
    _STORAGE_P_IN_ = 5e5
    _COLD_STORAGE_P_IN_ = 5e5
    _ENV_P_IN_ = 5e5
    _Q_DOT_MIN_ = 1e3  # and heat_flow rate (W)
    _D_T_SUPER_ = 5  # super heating of working fluid
    _D_T_MIN_ = 3.  # minimum approach temperature (pinch point)
    _COP_CHARGING = 3.144  # needed to calculate Q_env_discharging
    _T_REDUCTION_EVAP = -22  # if the curves cross in the evaporator this parameter may help
    _DT_COND_ = 7.
    # environment for heat transfer
    state_env_out = envFluid.set_state([_ENV_T_OUT_, _ENV_P_IN_], "TP")
    state_env_in = envFluid.set_state([_ENV_T_IN_, _ENV_P_IN_], "TP")

    # high T-storages
    state_sec_out = secFluid.set_state([_STORAGE_T_OUT_, _STORAGE_P_IN_], "TP")
    state_sec_in = secFluid.set_state([_STORAGE_T_IN_, _STORAGE_P_IN_], "TP")

    #  low T sorages:
    state_cold_out = coldFluid.set_state(
        [_COLD_STORAGE_T_OUT_, _COLD_STORAGE_P_IN_], "TP")
    state_cold_in = coldFluid.set_state(
        [_COLD_STORAGE_T_IN_, _COLD_STORAGE_P_IN_], "TP")

    # working fluid

    state_satv_evap = myFluid.set_state(
        [_STORAGE_T_IN_-_D_T_MIN_-_D_T_SUPER_+_T_REDUCTION_EVAP, 1.], "TQ")  # find high pressure
    p_high = state_satv_evap[1]

    T_OUT = _STORAGE_T_IN_ - _D_T_MIN_
    # Evaporator input comes from the pump-output

    state_out_evap = myFluid.set_state([p_high,
                                        T_OUT], "PT")
    # low pressure, condenser # BA 2023-11-14 three points needed: low, environment and slightly higher
    T_SATL = _COLD_STORAGE_T_IN_ + _D_T_MIN_ + _DT_COND_
    state_out_cond = myFluid.set_state([T_SATL, 0.], "TQ")  # find low pressure
    p_low = state_out_cond[1]
    # BA changed 2023-12-13 the fixed starting point for the cycle is the  fluid
    # state before the pump now.

    # the other states in the condenser are fixed by the expander outlet and
    # the Q_total : Q_low_stored ratio    eventually p_low must be varied until
    # the balance is fulfilled, since m_dot_w is fixed by the evaporator!

    FIXED_POINTS = {"eta_s": _ETA_S_,  # expander
                    "eta_s_p": _ETA_S_P_,  # pump
                    "p_low": p_low,
                    "p_high": p_high,
                    "T_hh": _STORAGE_T_IN_,
                    "h_h_out_sec": state_sec_out[2],
                    "h_h_out_w": state_out_evap[2],
                    "h_l_out_cold": state_cold_out[2],
                    "h_l_out_w": state_out_cond[2],
                    "h_env_in": state_env_in[2],
                    "h_env_out": state_env_out[2],
                    "T_hl": _STORAGE_T_OUT_,
                    "T_lh": _COLD_STORAGE_T_OUT_,
                    "T_ll": _COLD_STORAGE_T_IN_,  # 256.0,
                    "Q_dot_h": _Q_DOT_MIN_,
                    "d_temp_min": _D_T_MIN_,
                    "cop_charging": _COP_CHARGING  # needed to calculate Q_env_discharging
                    }

    orc0 = OrganicRankineCycle(
        [myFluid, secFluid, envFluid, coldFluid], FIXED_POINTS)
    eta_dis = orc0.calc_orc(True)
    print(f"eta(ORC: {eta_dis}")
    orc0.hp_plot()
